import collections.abc as cabc
import itertools

import pyxu.abc as pxa
import pyxu.info.deps as pxd
import pyxu.info.ptype as pxt
import pyxu.operator.linop.base as pxlb
import pyxu.operator.linop.diff as pxld
import pyxu.operator.linop.pad as pxlp
import pyxu.operator.linop.stencil.stencil as pxls
import pyxu.runtime as pxrt
import pyxu.util as pxu

try:
    import scipy.ndimage._filters as scif
except ImportError:
    import scipy.ndimage.filters as scif

import functools
import typing as typ

import numpy as np

IndexSpec = cabc.Sequence[pxt.Integer]
KernelSpec = pxls.Stencil.KernelSpec
ModeSpec = pxlp.Pad.ModeSpec


__all__ = [
    "MovingAverage",
    "Gaussian",
    "DifferenceOfGaussians",
    "DoG",
    "Laplace",
    "Sobel",
    "Prewitt",
    "Scharr",
    "StructureTensor",
]


def _to_canonical_form(_, arg_shape):
    if not isinstance(_, cabc.Sequence):
        _ = (_,) * len(arg_shape)
    else:
        assert len(_) == len(arg_shape)
    return _


def _get_axes(axis, ndim):
    if axis is None:
        axes = list(range(ndim))
    elif np.isscalar(axis):
        axes = [axis]
    else:
        axes = axis
    return axes


def _sanitize_inputs(arg_shape, dtype, gpu):
    ndim = len(arg_shape)

    if dtype is None:
        dtype = pxrt.getPrecision().value

    if gpu:
        assert pxd.CUPY_ENABLED
        import cupy as xp
    else:
        import numpy as xp
    return ndim, dtype, xp


def MovingAverage(
    arg_shape: pxt.NDArrayShape,
    size: typ.Union[typ.Tuple, pxt.Integer],
    center: typ.Optional[IndexSpec] = None,
    mode: ModeSpec = "constant",
    gpu: bool = False,
    dtype: typ.Optional[pxt.DType] = None,
):
    r"""
    Multi-dimensional moving average or uniform filter.

    Notes
    -----
    This operator performs a convolution between the input :math:`D`-dimensional NDArray :math:`\mathbf{x} \in
    \mathbb{R}^{N_0 \times \cdots \times N_{D-1}}` and a uniform :math:`D`-dimensional filter :math:`\mathbf{h} \in
    \mathbb{R}^{\text{size} \times \cdots \times \text{size}}` that computes the `size`-point local mean values using
    separable kernels for improved performance.

    .. math::

       y_{i} = \frac{1}{|\mathcal{N}_{i}|}\sum_{j \in \mathcal{N}_{i}} x_{j}

    Where :math:`\mathcal{N}_{i}` is the set of elements neighbouring the :math:`i`-th element of the input array, and
    :math:`\mathcal{N}_{i}` denotes its cardinality, i.e. the total number of neighbors.


    Parameters
    ----------
    arg_shape: NDArrayShape
        Shape of the input array.
    size: int, tuple
        Size of the moving average kernel.

        If a single integer value is provided, then the moving average filter will have as many dimensions as the input
        array.  If a tuple is provided, it should contain as many elements as `arg_shape`.  For example, the ``size=(1,
        3)`` will convolve the input image with the filter ``[[1, 1, 1]] / 3``.

    center: ~pyxu.operator.linop.filter.IndexSpec
        (i_1, ..., i_D) index of the kernel's center.

        `center` defines how a kernel is overlaid on inputs to produce outputs.
        For odd `size`, it defaults to the central element (``center=size//2``).
        For even `size` the desired center indices must be provided.

    mode: str, list[str]
        Boundary conditions.
        Multiple forms are accepted:

        * str: unique mode shared amongst dimensions.
          Must be one of:

          * 'constant' (default): zero-padding
          * 'wrap'
          * 'reflect'
          * 'symmetric'
          * 'edge'
        * tuple[str, ...]: the `d`-th dimension uses `mode[d]` as boundary condition.

        (See :py:func:`numpy.pad` for details.)
    gpu: bool
        Input NDArray type (`True` for GPU, `False` for CPU). Defaults to `False`.
    dtype: DType
        Working precision of the linear operator.

    Returns
    -------
    op: OpT
        MovingAverage

    Example
    -------

    .. plot::

       import matplotlib.pyplot as plt
       from pyxu.operator import MovingAverage

       arg_shape = (11, 11)
       image = np.zeros(arg_shape)
       image[5, 5] = 1.

       ma = MovingAverage(arg_shape, size=5)
       out = ma(image.ravel())
       plt.figure(figsize=(10, 5))
       plt.subplot(121)
       plt.imshow(image)
       plt.colorbar()
       plt.subplot(122)
       plt.imshow(out.reshape(*arg_shape))
       plt.colorbar()

    See Also
    --------
    :py:class:`~pyxu.operator.Gaussian`
    """
    size = _to_canonical_form(size, arg_shape)
    if center is None:
        assert all([s % 2 == 1 for s in size]), (
            "Can only infer center for odd `size`s. For even `size`s, please " "provide the desired `center`s."
        )
        center = [s // 2 for s in size]

    ndim, dtype, xp = _sanitize_inputs(arg_shape, dtype, gpu)

    kernel = [xp.ones(s, dtype=dtype) for s in size]  # use separable filters
    scale = 1 / np.prod(size)

    op = scale * pxls.Stencil(arg_shape=arg_shape, kernel=kernel, center=center, mode=mode)
    op._name = "MovingAverage"
    return op


def Gaussian(
    arg_shape: pxt.NDArrayShape,
    sigma: typ.Union[typ.Tuple[pxt.Real], pxt.Real] = 1.0,
    truncate: typ.Union[typ.Tuple[pxt.Real], pxt.Real] = 3.0,
    order: typ.Union[typ.Tuple[pxt.Integer], pxt.Integer] = 0,
    mode: ModeSpec = "constant",
    sampling: typ.Union[pxt.Real, cabc.Sequence[pxt.Real, ...]] = 1,
    gpu: bool = False,
    dtype: typ.Optional[pxt.DType] = None,
):
    r"""
    Multi-dimensional Gaussian filter.

    Notes
    -----
    This operator performs a convolution between the input :math:`D`-dimensional NDArray :math:`\mathbf{x} \in
    \mathbb{R}^{N_0 \times \cdots \times N_{D-1}}` and a Gaussian :math:`D`-dimensional filter :math:`\mathbf{h} \in
    \mathbb{R}^{\text{size} \times \cdots \times \text{size}}` using separable kernels for improved performance.

    .. math::

       y_{i} = \sum_{j \in \mathcal{N}_{i}} a_{j} x_{j} \exp(\frac{d_{ij}^{2}}{\sigma^{2}})

    Where :math:`\mathcal{N}_{i}` is the set of elements neighbouring the :math:`i`-th element of the input array, and
    :math:`a_{j} = \sum_{j \in \mathcal{N}_{i}} a_{j} \exp(\frac{d_{ij}^{2}}{\sigma^{2}})` normalizes the kernel to sum
    to one.


    Parameters
    ----------
    arg_shape: NDArrayShape
        Shape of the input array.
    sigma: float, tuple
        Standard deviation of the Gaussian kernel.

        If a scalar value is provided, then the Gaussian filter will have as many dimensions as the input array.
        If a tuple is provided, it should contain as many elements as `arg_shape`.
        Use ``0`` to prevent filtering in a given dimension.
        For example, the ``sigma=(0, 3)`` will convolve the input image in its last dimension.
    truncate: float, tuple
        Truncate the filter at this many standard deviations.
        Defaults to 3.0.
    order: int, tuple
        Gaussian derivative order.

        Use ``0`` for the standard Gaussian kernel.
    mode: str, list[str]
        Boundary conditions.
        Multiple forms are accepted:

        * str: unique mode shared amongst dimensions.
          Must be one of:

          * 'constant' (default): zero-padding
          * 'wrap'
          * 'reflect'
          * 'symmetric'
          * 'edge'
        * tuple[str, ...]: the `d`-th dimension uses `mode[d]` as boundary condition.

        (See :py:func:`numpy.pad` for details.)
    sampling: Real, list[Real]
        Sampling step (i.e. distance between two consecutive elements of an array).
        Defaults to 1.
    gpu: bool
        Input NDArray type (`True` for GPU, `False` for CPU).
        Defaults to `False`.
    dtype: DType
        Working precision of the linear operator.

    Returns
    -------
    op: OpT
        Gaussian

    Example
    -------

    .. plot::

       import matplotlib.pyplot as plt
       from pyxu.operator import Gaussian

       arg_shape = (11, 11)
       image = np.zeros(arg_shape)
       image[5, 5] = 1.

       gaussian = Gaussian(arg_shape, sigma=3)
       out = gaussian(image.ravel())
       plt.figure(figsize=(10, 5))
       plt.subplot(121)
       plt.imshow(image)
       plt.colorbar()
       plt.subplot(122)
       plt.imshow(out.reshape(*arg_shape))
       plt.colorbar()

    See Also
    --------
    :py:class:`~pyxu.operator.MovingAverage`,
    :py:class:`~pyxu.operator.DifferenceOfGaussians`
    """

    ndim, dtype, xp = _sanitize_inputs(arg_shape, dtype, gpu)
    sigma = _to_canonical_form(sigma, arg_shape)
    truncate = _to_canonical_form(truncate, arg_shape)
    order = _to_canonical_form(order, arg_shape)
    sampling = _to_canonical_form(sampling, arg_shape)

    kernel = [
        xp.array([1], dtype=dtype),
    ] * len(arg_shape)
    center = [0 for _ in range(len(arg_shape))]

    for i, (sigma_, truncate_, order_, sampling_) in enumerate(zip(sigma, truncate, order, sampling)):
        if sigma_:
            sigma_pix = sigma_ / sampling_
            radius = int(truncate_ * float(sigma_pix) + 0.5)
            kernel[i] = xp.asarray(np.flip(scif._gaussian_kernel1d(sigma_pix, order_, radius)), dtype=dtype)
            kernel[i] /= sampling_**order_
            center[i] = radius

    op = pxls.Stencil(arg_shape=arg_shape, kernel=kernel, center=center, mode=mode)
    op._name = "Gaussian"
    return op


def DifferenceOfGaussians(
    arg_shape: pxt.NDArrayShape,
    low_sigma=1.0,
    high_sigma=None,
    low_truncate=3.0,
    high_truncate=3.0,
    mode: ModeSpec = "constant",
    sampling: typ.Union[pxt.Real, cabc.Sequence[pxt.Real, ...]] = 1,
    gpu: bool = False,
    dtype: typ.Optional[pxt.DType] = None,
):
    r"""
    Multi-dimensional Difference of Gaussians filter.

    Notes
    -----

    This operator uses the Difference of Gaussians (DoG) method to a :math:`D`-dimensional NDArray :math:`\mathbf{x} \in
    \mathbb{R}^{N_0 \times \cdots \times N_{D-1}}` using separable kernels for improved performance.
    The DoG method blurs the input image with two Gaussian kernels with different sigma, and subtracts the more-blurred
    signal from the less-blurred image.
    This creates an output signal containing only the information from the original signal at the spatial scale
    indicated by the two sigmas.

    Parameters
    ----------
    arg_shape: NDArrayShape
        Shape of the input array.
    low_sigma: float, tuple
        Standard deviation of the Gaussian kernel with smaller sigmas across all axes.

        If a scalar value is provided, then the Gaussian filter will have as many dimensions as the input array.
        If a tuple is provided, it should contain as many elements as `arg_shape`.
        Use ``0`` to prevent filtering in a given dimension.
        For example, the ``low_sigma=(0, 3)`` will convolve the input image in its last dimension.
    high_sigma: float, tuple, None
        Standard deviation of the Gaussian kernel with larger sigmas across all axes.
        If ``None`` is given (default), sigmas for all axes are calculated as ``1.6 * low_sigma``.
    low_truncate: float, tuple
        Truncate the filter at this many standard deviations.
        Defaults to 3.0.
    high_truncate: float, tuple
        Truncate the filter at this many standard deviations.
        Defaults to 3.0.
    mode: str, list[str]
        Boundary conditions.
        Multiple forms are accepted:

        * str: unique mode shared amongst dimensions.
          Must be one of:

          * 'constant' (default): zero-padding
          * 'wrap'
          * 'reflect'
          * 'symmetric'
          * 'edge'
        * tuple[str, ...]: the `d`-th dimension uses `mode[d]` as boundary condition.

        (See :py:func:`numpy.pad` for details.)
    sampling: Real, list[Real]
        Sampling step (i.e. distance between two consecutive elements of an array).
        Defaults to 1.
    gpu: bool
        Input NDArray type (`True` for GPU, `False` for CPU). Defaults to `False`.
    dtype: DType
        Working precision of the linear operator.

    Returns
    -------
    op: OpT
        DifferenceOfGaussians

    Example
    -------

    .. plot::

       import matplotlib.pyplot as plt
       from pyxu.operator import DoG

       arg_shape = (11, 11)
       image = np.zeros(arg_shape)
       image[5, 5] = 1.

       dog = DoG(arg_shape, low_sigma=3)
       out = dog(image.ravel())
       plt.figure(figsize=(10, 5))
       plt.subplot(121)
       plt.imshow(image)
       plt.colorbar()
       plt.subplot(122)
       plt.imshow(out.reshape(*arg_shape))
       plt.colorbar()

    See Also
    --------
    :py:class:`~pyxu.operator.Gaussian`,
    :py:class:`~pyxu.operator.Sobel`,
    :py:class:`~pyxu.operator.Prewitt`,
    :py:class:`~pyxu.operator.Scharr`,
    :py:class:`~pyxu.operator.StructureTensor`
    """

    low_sigma = _to_canonical_form(low_sigma, arg_shape)
    if high_sigma is None:
        high_sigma = tuple(s * 1.6 for s in low_sigma)

    high_sigma = _to_canonical_form(high_sigma, arg_shape)
    low_truncate = _to_canonical_form(low_truncate, arg_shape)
    high_truncate = _to_canonical_form(high_truncate, arg_shape)

    kwargs = {
        "arg_shape": arg_shape,
        "order": 0,
        "mode": mode,
        "gpu": gpu,
        "dtype": dtype,
        "sampling": sampling,
    }
    op_low = Gaussian(sigma=low_sigma, truncate=low_truncate, **kwargs)
    op_high = Gaussian(sigma=high_sigma, truncate=high_truncate, **kwargs)
    op = op_low - op_high
    op._name = "DifferenceOfGaussians"
    return op


DoG = DifferenceOfGaussians  #: Alias of :py:func:`~pyxu.operator.DifferenceOfGaussians`.


def Laplace(
    arg_shape: pxt.NDArrayShape,
    mode: ModeSpec = "constant",
    sampling: typ.Union[pxt.Real, cabc.Sequence[pxt.Real, ...]] = 1,
    gpu: bool = False,
    dtype: typ.Optional[pxt.DType] = None,
):
    r"""
    Multi-dimensional Laplace filter.

    The implementation is based on second derivatives approximated via finite differences.

    Notes
    -----
    This operator uses the applies the Laplace kernel :math:`[1 -2 1]` to a :math:`D`-dimensional NDArray
    :math:`\mathbf{x} \in \mathbb{R}^{N_0 \times \cdots \times N_{D-1}}` using separable kernels for improved
    performance.  The Laplace filter is commonly used to find high-frequency components in the signal, such as for
    example, the edges in an image.

    Parameters
    ----------
    arg_shape: NDArrayShape
        Shape of the input array.
    mode: str, list[str]
        Boundary conditions.
        Multiple forms are accepted:

        * str: unique mode shared amongst dimensions.
          Must be one of:

          * 'constant' (default): zero-padding
          * 'wrap'
          * 'reflect'
          * 'symmetric'
          * 'edge'
        * tuple[str, ...]: the `d`-th dimension uses `mode[d]` as boundary condition.

        (See :py:func:`numpy.pad` for details.)
    sampling: Real, list[Real]
            Sampling step (i.e. distance between two consecutive elements of an array).
            Defaults to 1.
    gpu: bool
        Input NDArray type (`True` for GPU, `False` for CPU). Defaults to `False`.
    dtype: DType
        Working precision of the linear operator.

    Returns
    -------
    op: OpT
        DifferenceOfGaussians

    Example
    -------

    .. plot::

       import matplotlib.pyplot as plt
       from pyxu.operator import Laplace

       arg_shape = (11, 11)
       image = np.zeros(arg_shape)
       image[5, 5] = 1.

       laplace = Laplace(arg_shape)
       out = laplace(image.ravel())
       plt.figure(figsize=(10, 5))
       plt.subplot(121)
       plt.imshow(image)
       plt.colorbar()
       plt.subplot(122)
       plt.imshow(out.reshape(*arg_shape))
       plt.colorbar()

    See Also
    --------
    :py:class:`~pyxu.operator.Sobel`,
    :py:class:`~pyxu.operator.Prewitt`,
    :py:class:`~pyxu.operator.Scharr`,
    """

    ndim, dtype, xp = _sanitize_inputs(arg_shape, dtype, gpu)
    sampling = _to_canonical_form(sampling, arg_shape)
    centers = [[1 if i == dim else 0 for i in range(ndim)] for dim in range(ndim)]
    kernels = [
        xp.array([1.0, -2.0, 1.0]).reshape([-1 if i == dim else 1 for i in range(ndim)]) / sampling[dim]
        for dim in range(ndim)
    ]
    ops = [pxls.Stencil(arg_shape=arg_shape, kernel=k, center=c, mode=mode) for (k, c) in zip(kernels, centers)]
    op = functools.reduce(lambda x, y: x + y, ops)
    op._name = "Laplace"
    return op


def Sobel(
    arg_shape: pxt.NDArrayShape,
    axis: typ.Optional[typ.Tuple] = None,
    mode: ModeSpec = "constant",
    sampling: typ.Union[pxt.Real, cabc.Sequence[pxt.Real, ...]] = 1,
    gpu: bool = False,
    dtype: typ.Optional[pxt.DType] = None,
):
    r"""
    Multi-dimensional Sobel filter.

    Notes
    -----

    This operator uses the applies the multi-dimensional Sobel filter to a :math:`D`-dimensional NDArray
    :math:`\mathbf{x} \in \mathbb{R}^{N_0 \times \cdots \times N_{D-1}}` using separable kernels for improved
    performance.  The Sobel filter applies the following edge filter in the dimensions of interest: ``[1, 0, -1]`` and
    the smoothing filter on the rest of dimensions: ``[1, 2, 1] / 4``.  The Sobel filter is commonly used to find
    high-frequency components in the signal, such as for example, the edges in an image.

    Parameters
    ----------
    arg_shape: NDArrayShape
        Shape of the input array.
    axis: int, tuple
        Compute the edge filter along this axis. If not provided, the edge magnitude is computed.

        This is defined as: ``np.sqrt(sum([sobel(array, axis=i)**2 for i in range(array.ndim)]) / array.ndim)`` The
        magnitude is also computed if axis is a sequence.

    mode: str, list[str]
        Boundary conditions.
        Multiple forms are accepted:

        * str: unique mode shared amongst dimensions.
          Must be one of:

          * 'constant' (default): zero-padding
          * 'wrap'
          * 'reflect'
          * 'symmetric'
          * 'edge'
        * tuple[str, ...]: the `d`-th dimension uses `mode[d]` as boundary condition.

        (See :py:func:`numpy.pad` for details.)
    sampling: Real, list[Real]
            Sampling step (i.e. distance between two consecutive elements of an array).
            Defaults to 1.
    gpu: bool
        Input NDArray type (`True` for GPU, `False` for CPU). Defaults to `False`.
    dtype: DType
        Working precision of the linear operator.

    Returns
    -------
    op: OpT
        Sobel

    Example
    -------

    .. plot::

       import matplotlib.pyplot as plt
       from pyxu.operator import Sobel

       arg_shape = (11, 11)
       image = np.zeros(arg_shape)
       image[5, 5] = 1.

       sobel = Sobel(arg_shape)
       out = sobel(image.ravel())
       plt.figure(figsize=(10, 5))
       plt.subplot(121)
       plt.imshow(image)
       plt.colorbar()
       plt.subplot(122)
       plt.imshow(out.reshape(*arg_shape))
       plt.colorbar()

    See Also
    --------
    :py:class:`~pyxu.operator.Prewitt`,
    :py:class:`~pyxu.operator.Scharr`,
    """
    smooth_kernel = np.array([1, 2, 1]) / 4
    return _EdgeFilter(
        arg_shape=arg_shape,
        smooth_kernel=smooth_kernel,
        filter_name="SobelFilter",
        axis=axis,
        mode=mode,
        sampling=sampling,
        gpu=gpu,
        dtype=dtype,
    )


def Prewitt(
    arg_shape: pxt.NDArrayShape,
    axis: typ.Optional[typ.Tuple] = None,
    mode: ModeSpec = "constant",
    sampling: typ.Union[pxt.Real, cabc.Sequence[pxt.Real, ...]] = 1,
    gpu: bool = False,
    dtype: typ.Optional[pxt.DType] = None,
):
    r"""
    Multi-dimensional Prewitt filter.

    Notes
    -----

    This operator uses the applies the multi-dimensional Prewitt filter to a :math:`D`-dimensional NDArray
    :math:`\mathbf{x} \in \mathbb{R}^{N_0 \times \cdots \times N_{D-1}}` using separable kernels for improved
    performance.  The Prewitt filter applies the following edge filter in the dimensions of interest: ``[1, 0, -1]``,
    and the smoothing filter on the rest of dimensions: ``[1, 1, 1] / 3``.  The Prewitt filter is commonly used to find
    high-frequency components in the signal, such as for example, the edges in an image.

    Parameters
    ----------
    arg_shape: NDArrayShape
        Shape of the input array.
    axis: int, tuple
        Compute the edge filter along this axis. If not provided, the edge magnitude is computed. This is defined as:

        ``np.sqrt(sum([prewitt(array, axis=i)**2 for i in range(array.ndim)]) / array.ndim)`` The magnitude is also
        computed if axis is a sequence.

    mode: str, list[str]
        Boundary conditions.
        Multiple forms are accepted:

        * str: unique mode shared amongst dimensions.
          Must be one of:

          * 'constant' (default): zero-padding
          * 'wrap'
          * 'reflect'
          * 'symmetric'
          * 'edge'
        * tuple[str, ...]: the `d`-th dimension uses `mode[d]` as boundary condition.

        (See :py:func:`numpy.pad` for details.)
    sampling: Real, list[Real]
            Sampling step (i.e. distance between two consecutive elements of an array).
            Defaults to 1.
    gpu: bool
        Input NDArray type (`True` for GPU, `False` for CPU). Defaults to `False`.
    dtype: DType
        Working precision of the linear operator.

    Returns
    -------
    op: OpT
        Prewitt

    Example
    -------

    .. plot::

       import matplotlib.pyplot as plt
       from pyxu.operator import Prewitt

       arg_shape = (11, 11)
       image = np.zeros(arg_shape)
       image[5, 5] = 1.

       prewitt = Prewitt(arg_shape)
       out = prewitt(image.ravel())
       plt.figure(figsize=(10, 5))
       plt.subplot(121)
       plt.imshow(image)
       plt.colorbar()
       plt.subplot(122)
       plt.imshow(out.reshape(*arg_shape))
       plt.colorbar()

    See Also
    --------
    :py:class:`~pyxu.operator.Sobel`,
    :py:class:`~pyxu.operator.Scharr`,
    """
    smooth_kernel = np.full((3,), 1 / 3)
    return _EdgeFilter(
        arg_shape=arg_shape,
        smooth_kernel=smooth_kernel,
        filter_name="Prewitt",
        axis=axis,
        mode=mode,
        sampling=sampling,
        gpu=gpu,
        dtype=dtype,
    )


def Scharr(
    arg_shape: pxt.NDArrayShape,
    axis: typ.Optional[typ.Tuple] = None,
    mode: ModeSpec = "constant",
    sampling: typ.Union[pxt.Real, cabc.Sequence[pxt.Real, ...]] = 1,
    gpu: bool = False,
    dtype: typ.Optional[pxt.DType] = None,
):
    r"""
    Multi-dimensional Scharr filter.

    Notes
    -----

    This operator uses the applies the multi-dimensional Scharr filter to a :math:`D`-dimensional NDArray
    :math:`\mathbf{x} \in \mathbb{R}^{N_0 \times \cdots \times N_{D-1}}` using separable kernels for improved
    performance.  The Scharr filter applies the following edge filter in the dimensions of interest: ``[1, 0, -1]``, and
    the smoothing filter on the rest of dimensions: ``[3, 10, 3] / 16``.  The Scharr filter is commonly used to find
    high-frequency components in the signal, such as for example, the edges in an image.

    Parameters
    ----------
    arg_shape: NDArrayShape
        Shape of the input array.
    axis: int, tuple
        Compute the edge filter along this axis. If not provided, the edge magnitude is computed. This is defined as:

        ``np.sqrt(sum([scharr(array, axis=i)**2 for i in range(array.ndim)]) / array.ndim)`` The magnitude is also
        computed if axis is a sequence.
    mode: str, list[str]
        Boundary conditions.
        Multiple forms are accepted:

        * str: unique mode shared amongst dimensions.
          Must be one of:

          * 'constant' (default): zero-padding
          * 'wrap'
          * 'reflect'
          * 'symmetric'
          * 'edge'
        * tuple[str, ...]: the `d`-th dimension uses `mode[d]` as boundary condition.

        (See :py:func:`numpy.pad` for details.)
    sampling: Real, list[Real]
            Sampling step (i.e. distance between two consecutive elements of an array).
            Defaults to 1.
    gpu: bool
        Input NDArray type (`True` for GPU, `False` for CPU). Defaults to `False`.
    dtype: DType
        Working precision of the linear operator.

    Returns
    -------
    op: OpT
        Scharr

    Example
    -------

    .. plot::

       import matplotlib.pyplot as plt
       from pyxu.operator import Scharr

       arg_shape = (11, 11)
       image = np.zeros(arg_shape)
       image[5, 5] = 1.

       scharr = Scharr(arg_shape)
       out = scharr(image.ravel())
       plt.figure(figsize=(10, 5))
       plt.subplot(121)
       plt.imshow(image)
       plt.colorbar()
       plt.subplot(122)
       plt.imshow(out.reshape(*arg_shape))
       plt.colorbar()

    See Also
    --------
    :py:class:`~pyxu.operator.Sobel`,
    :py:class:`~pyxu.operator.Prewitt`,
    """
    smooth_kernel = np.array([3, 10, 3]) / 16
    return _EdgeFilter(
        arg_shape=arg_shape,
        smooth_kernel=smooth_kernel,
        filter_name="Scharr",
        axis=axis,
        mode=mode,
        sampling=sampling,
        gpu=gpu,
        dtype=dtype,
    )


def _EdgeFilter(
    arg_shape: pxt.NDArrayShape,
    smooth_kernel: KernelSpec,
    filter_name: str,
    axis: typ.Optional[typ.Tuple] = None,
    mode: ModeSpec = "constant",
    sampling: typ.Union[pxt.Real, cabc.Sequence[pxt.Real, ...]] = 1,
    gpu: bool = False,
    dtype: typ.Optional[pxt.DType] = None,
):
    from pyxu.operator.map.ufunc import sqrt, square

    ndim, dtype, xp = _sanitize_inputs(arg_shape, dtype, gpu)
    sampling = _to_canonical_form(sampling, arg_shape)

    axes = _get_axes(axis, ndim)

    return_magnitude = len(axes) > 1

    op_list = []
    for edge_dim in axes:
        kernel = [xp.array(1)] * len(arg_shape)
        center = np.ones(len(arg_shape), dtype=int)
        # We define the kernel reversed compared to Scipy or Skimage because we use correlation instead of convolution
        kernel[edge_dim] = xp.array([-1, 0, 1], dtype=dtype) / sampling[edge_dim]
        smooth_axes = list(set(range(ndim)) - {edge_dim})
        for smooth_dim in smooth_axes:
            kernel[smooth_dim] = xp.asarray(smooth_kernel, dtype=dtype) / sampling[smooth_dim]

        if return_magnitude:
            op_list.append(square(pxls.Stencil(arg_shape=arg_shape, kernel=kernel, center=center, mode=mode)))
        else:
            op_list.append(pxls.Stencil(arg_shape=arg_shape, kernel=kernel, center=center, mode=mode))

    op = functools.reduce(lambda x, y: x + y, op_list)
    if return_magnitude:
        op = (1 / np.sqrt(ndim)) * sqrt(op)

    op._name = filter_name
    return op


class StructureTensor(pxa.DiffMap):
    r"""
    Structure tensor operator.

    Notes
    -----
    The Structure Tensor, also known as the second-order moment tensor or the inertia tensor, is a matrix derived from
    the gradient of a function. It describes the distribution of the gradient (i.e., its prominent directions) in a
    specified neighbourhood around a point, and the degree to which those directions are coherent.  The structure tensor
    of a :math:`D`-dimensional signal :math:`\mathbf{f} \in \mathbb{R}^{N_0 \times \cdots \times N_{D-1}}` can be
    written as:

    .. math::

        \mathbf{S}_\sigma \mathbf{f} = \mathbf{g}_{\sigma} * \nabla\mathbf{f} (\nabla\mathbf{f})^{\top} = \mathbf{g}_{\sigma} *
        \begin{bmatrix}
        \left( \dfrac{ \partial\mathbf{f} }{ \partial x_{0} } \right)^2 &  \dfrac{ \partial^{2}\mathbf{f} }{ \partial x_{0}\,\partial x_{1} } & \cdots & \dfrac{ \partial\mathbf{f} }{ \partial x_{0} } \dfrac{ \partial\mathbf{f} }{ \partial x_{D-1} } \\
        \dfrac{ \partial\mathbf{f} }{ \partial x_{1} } \dfrac{ \partial\mathbf{f} }{ \partial x_{0} } & \left( \dfrac{ \partial\mathbf{f} }{ \partial x_{1} }\right)^2 & \cdots & \dfrac{ \partial\mathbf{f} }{ \partial x_{1} } \dfrac{ \partial\mathbf{f} }{ \partial x_{D-1} } \\
        \vdots & \vdots & \ddots & \vdots \\
        \dfrac{ \partial\mathbf{f} }{ \partial x_{D-1} } \dfrac{ \partial\mathbf{f} }{ \partial x_{0} } & \dfrac{ \partial\mathbf{f} }{ \partial x_{D-1} } \dfrac{ \partial\mathbf{f} }{ \partial x_{1} } & \cdots & \left( \dfrac{ \partial\mathbf{f} }{ \partial x_{D-1}} \right)^2
        \end{bmatrix},

    where :math:`\mathbf{g}_{\sigma} \in \mathbb{R}^{N_0 \times \cdots \times N_{D-1}}` is a discrete Gaussian filter
    with standard variation :math:`\sigma` with which a convolution is performed elementwise.

    However, due to the symmetry of the structure tensor, only the upper triangular part is computed in practice:

    .. math::

        \mathbf{H}_{\mathbf{v}_1, \ldots ,\mathbf{v}_m} \mathbf{f} = \mathbf{g}_{\sigma} * \begin{bmatrix}
        \left( \dfrac{ \partial\mathbf{f} }{ \partial x_{0} } \right)^2 \\
        \dfrac{ \partial^{2}\mathbf{f} }{ \partial x_{0}\,\partial x_{1} } \\
        \vdots \\
        \left( \dfrac{ \partial\mathbf{f} }{ \partial x_{D-1}} \right)^2
        \end{bmatrix} \mathbf{f} \in \mathbb{R}^{\frac{D (D-1)}{2} \times N_0 \times \cdots \times N_{D-1}}

    Remark
    ------
    In case of using the finite differences (`diff_type="fd"`), the finite difference scheme defaults to `central` (see
    :py:class:`~pyxu.operator.PartialDerivative`).

    Example
    -------

    .. plot::

       import numpy as np
       import matplotlib.pyplot as plt
       from pyxu.operator import StructureTensor
       from pyxu.util.misc import peaks

       # Define input image
       n = 1000
       x = np.linspace(-3, 3, n)
       xx, yy = np.meshgrid(x, x)
       image = peaks(xx, yy)
       nsamples = 2
       arg_shape = image.shape  # (1000, 1000)
       images = np.tile(image, (nsamples, 1, 1)).reshape(nsamples, -1)
       print(images.shape)  # (2, 1000000)
       # Instantiate structure tensor operator
       structure_tensor = StructureTensor(arg_shape=arg_shape)

       outputs = structure_tensor(images)
       print(outputs.shape)  # (2, 3000000)
       # Plot
       outputs = structure_tensor.unravel(outputs)
       print(outputs.shape)  # (2, 3, 1000, 1000)
       plt.figure()
       plt.imshow(images[0].reshape(arg_shape))
       plt.colorbar()
       plt.title("Image")
       plt.axis("off")

       plt.figure()
       plt.imshow(outputs[0][0].reshape(arg_shape))
       plt.colorbar()
       plt.title(r"$\hat{S}_{xx}$")
       plt.axis("off")

       plt.figure()
       plt.imshow(outputs[0][1].reshape(arg_shape))
       plt.colorbar()
       plt.title(r"$\hat{S}_{xy}$")
       plt.axis("off")

       plt.figure()
       plt.imshow(outputs[0][2].reshape(arg_shape))
       plt.colorbar()
       plt.title(r"$\hat{S}_{yy}$")
       plt.axis("off")

    See Also
    --------
    :py:class:`~pyxu.operator.PartialDerivative`,
    :py:class:`~pyxu.operator.Gradient`,
    :py:class:`~pyxu.operator.Hessian`
    """

    def __init__(
        self,
        arg_shape: pxt.NDArrayShape,
        diff_method="fd",
        smooth_sigma: typ.Union[pxt.Real, tuple[pxt.Real, ...]] = 1.0,
        smooth_truncate: typ.Union[pxt.Real, tuple[pxt.Real, ...]] = 3.0,
        mode: ModeSpec = "constant",
        sampling: typ.Union[pxt.Real, tuple[pxt.Real, ...]] = 1,
        gpu: bool = False,
        dtype: typ.Optional[pxt.DType] = None,
        parallel: bool = False,
        **diff_kwargs,
    ):
        self.arg_shape = arg_shape
        size = int(np.prod(arg_shape))
        ndim = len(arg_shape)
        ntriu = (ndim * (ndim + 1)) // 2
        super().__init__(shape=(ntriu * size, size))
        self.directions = tuple(
            list(_) for _ in itertools.combinations_with_replacement(np.arange(len(arg_shape)).astype(int), 2)
        )

        if diff_method == "fd":
            diff_kwargs.update({"scheme": diff_kwargs.pop("scheme", "central")})
        self.grad = pxld.Gradient(
            arg_shape=arg_shape,
            directions=None,
            mode=mode,
            gpu=gpu,
            dtype=dtype,
            sampling=sampling,
            parallel=parallel,
            **diff_kwargs,
        )

        if smooth_sigma:
            self.smooth = Gaussian(
                arg_shape=arg_shape,
                sigma=smooth_sigma,
                truncate=smooth_truncate,
                order=0,
                mode=mode,
                sampling=sampling,
                gpu=gpu,
                dtype=dtype,
            )
        else:
            self.smooth = pxlb.IdentityOp(dim=np.prod(arg_shape).item())

    def unravel(self, arr):
        return arr.reshape(*arr.shape[:-1], -1, *self.arg_shape)

    def ravel(self, arr):
        return arr.reshape(*arr.shape[: -1 - len(self.arg_shape)], -1)

    def apply(self, arr):
        xp = pxu.get_array_module(arr)
        sh = arr.shape[:-1]
        grad = self.grad.unravel(self.grad(arr))

        def slice_(ax):
            return (slice(None),) * len(sh) + (slice(ax, ax + 1),)

        return xp.concatenate(
            [
                self.grad.unravel(self.smooth((grad[slice_(i)] * grad[slice_(j)]).reshape(*sh, -1)))
                for i, j in self.directions
            ],
            axis=len(sh),
        ).reshape(*sh, -1)



import itertools

import numpy as np
import pytest

import pyxu.abc as pxa
import pyxu.info.deps as pxd
import pyxu.info.ptype as pxt
import pyxu.operator as pxo
import pyxu.runtime as pxrt
import pyxu_tests.operator.conftest as conftest


class TestL1Ball(conftest.ProxFuncT):
    @pytest.fixture(
        params=itertools.product(
            ((5, pxo.L1Ball(dim=5, radius=1)),),  # dim, op
            pxd.NDArrayInfo,
            pxrt.Width,
        )
    )
    def _spec(self, request):
        return request.param

    @pytest.fixture
    def spec(self, _spec):
        return _spec[0][1], _spec[1], _spec[2]

    @pytest.fixture
    def dim(self, _spec):
        return _spec[0][0]

    @pytest.fixture
    def data_shape(self, dim):
        return (1, dim)

    @pytest.fixture(
        params=[  # 2 evaluation points
            dict(
                in_=dict(arr=np.zeros((5,))),
                out=np.zeros((1,)),
            ),
            dict(
                in_=dict(arr=np.arange(-3, 2)),
                out=np.array([np.inf]),
            ),
        ]
    )
    def data_apply(self, request):
        return request.param

    @pytest.fixture(
        params=[  # 2 evaluation points
            dict(
                in_=dict(
                    arr=np.zeros((5,)),
                    tau=1,
                ),
                out=np.zeros((5,)),
            ),
            dict(
                in_=dict(
                    arr=np.arange(-3, 2),
                    tau=1,
                ),
                out=np.array([-1, 0, 0, 0, 0]),
            ),
        ]
    )
    def data_prox(self, request):
        return request.param

    @pytest.fixture
    def data_math_lipschitz(self, dim):
        N_test, dim = 10, self._sanitize(dim, 3)
        return self._random_array((N_test, dim))


class TestL2Ball(conftest.ProxFuncT):
    @pytest.fixture(
        params=itertools.product(
            ((5, pxo.L2Ball(dim=5, radius=1)),),  # dim, op
            pxd.NDArrayInfo,
            pxrt.Width,
        )
    )
    def _spec(self, request):
        return request.param

    @pytest.fixture
    def spec(self, _spec):
        return _spec[0][1], _spec[1], _spec[2]

    @pytest.fixture
    def dim(self, _spec):
        return _spec[0][0]

    @pytest.fixture
    def data_shape(self, dim):
        return (1, dim)

    @pytest.fixture(
        params=[  # 2 evaluation points
            dict(
                in_=dict(arr=np.zeros((5,))),
                out=np.zeros((1,)),
            ),
            dict(
                in_=dict(arr=np.arange(-3, 2)),
                out=np.array([np.inf]),
            ),
        ]
    )
    def data_apply(self, request):
        return request.param

    @pytest.fixture(
        params=[  # 2 evaluation points
            dict(
                in_=dict(
                    arr=np.zeros((5,)),
                    tau=1,
                ),
                out=np.zeros((5,)),
            ),
            dict(
                in_=dict(
                    arr=np.arange(-3, 2),
                    tau=1,
                ),
                out=np.arange(-3, 2) / np.linalg.norm(np.arange(-3, 2)),
            ),
        ]
    )
    def data_prox(self, request):
        return request.param

    @pytest.fixture
    def data_math_lipschitz(self, dim):
        N_test, dim = 10, self._sanitize(dim, 3)
        return self._random_array((N_test, dim))


class TestLInfinityBall(conftest.ProxFuncT):
    @pytest.fixture(
        params=itertools.product(
            ((5, pxo.LInfinityBall(dim=5, radius=1)),),  # dim, op
            pxd.NDArrayInfo,
            pxrt.Width,
        )
    )
    def _spec(self, request):
        return request.param

    @pytest.fixture
    def spec(self, _spec):
        return _spec[0][1], _spec[1], _spec[2]

    @pytest.fixture
    def dim(self, _spec):
        return _spec[0][0]

    @pytest.fixture
    def data_shape(self, dim):
        return (1, dim)

    @pytest.fixture(
        params=[  # 2 evaluation points
            dict(
                in_=dict(arr=np.zeros((5,))),
                out=np.zeros((1,)),
            ),
            dict(
                in_=dict(arr=np.arange(-3, 2)),
                out=np.array([np.inf]),
            ),
        ]
    )
    def data_apply(self, request):
        return request.param

    @pytest.fixture(
        params=[  # 2 evaluation points
            dict(
                in_=dict(
                    arr=np.zeros((5,)),
                    tau=1,
                ),
                out=np.zeros((5,)),
            ),
            dict(
                in_=dict(
                    arr=np.arange(-3, 2),
                    tau=1,
                ),
                out=np.array([-1, -1, -1, 0, 1]),
            ),
        ]
    )
    def data_prox(self, request):
        return request.param

    @pytest.fixture
    def data_math_lipschitz(self, dim):
        N_test, dim = 10, self._sanitize(dim, 3)
        return self._random_array((N_test, dim))


class TestPositiveOrthant(conftest.ProxFuncT):
    @pytest.fixture(
        params=itertools.product(
            ((5, pxo.PositiveOrthant(dim=5)),),  # dim, op
            pxd.NDArrayInfo,
            pxrt.Width,
        )
    )
    def _spec(self, request):
        return request.param

    @pytest.fixture
    def spec(self, _spec):
        return _spec[0][1], _spec[1], _spec[2]

    @pytest.fixture
    def dim(self, _spec):
        return _spec[0][0]

    @pytest.fixture
    def data_shape(self, dim):
        return (1, dim)

    @pytest.fixture(
        params=[  # 2 evaluation points
            dict(
                in_=dict(arr=np.zeros((5,))),
                out=np.zeros((1,)),
            ),
            dict(
                in_=dict(arr=np.arange(-3, 2)),
                out=np.array([np.inf]),
            ),
        ]
    )
    def data_apply(self, request):
        return request.param

    @pytest.fixture(
        params=[  # 2 evaluation points
            dict(
                in_=dict(
                    arr=np.zeros((5,)),
                    tau=1,
                ),
                out=np.zeros((5,)),
            ),
            dict(
                in_=dict(
                    arr=np.arange(-3, 2),
                    tau=1,
                ),
                out=np.array([0, 0, 0, 0, 1]),
            ),
        ]
    )
    def data_prox(self, request):
        return request.param

    @pytest.fixture
    def data_math_lipschitz(self, dim):
        N_test, dim = 10, self._sanitize(dim, 3)
        return self._random_array((N_test, dim))


class TestHyperSlab(conftest.ProxFuncT):
    @pytest.fixture(
        params=itertools.product(
            pxd.NDArrayInfo,
            pxrt.Width,
        )
    )
    def spec(self, request) -> tuple[pxt.OpT, pxd.NDArrayInfo, pxrt.Width]:
        ndi, width = request.param
        self._skip_if_unsupported(ndi)
        xp = ndi.module()

        v = xp.array([1, 1], dtype=width.value)
        a = pxa.LinFunc.from_array(v, enable_warnings=False)
        op = pxo.HyperSlab(a, lb=-1, ub=2)

        return op, ndi, width

    @pytest.fixture
    def data_shape(self) -> pxt.OpShape:
        return (1, 2)

    @pytest.fixture(
        params=[
            (np.r_[0, 0], np.r_[0]),
            (np.r_[2, 0], np.r_[0]),
            (np.r_[-1, 0], np.r_[0]),
            (np.r_[0, 2], np.r_[0]),
            (np.r_[0, -1], np.r_[0]),
            (np.r_[1.05, 1.05], np.r_[np.inf]),
            (np.r_[-0.5, -0.6], np.r_[np.inf]),
        ]
    )
    def data_apply(self, request) -> conftest.DataLike:
        arr, out = request.param
        return dict(
            in_=dict(arr=arr),
            out=out,
        )

    @pytest.fixture(
        params=[
            (np.r_[0, 0], np.r_[0, 0]),
            (np.r_[2.1, 3], np.r_[0.55, 1.45]),
            (np.r_[-3, -2], np.r_[-1, 0]),
            (np.r_[1, 0.5], np.r_[1, 0.5]),
        ]
    )
    def data_prox(self, request) -> conftest.DataLike:
        arr, out = request.param
        return dict(
            in_=dict(
                arr=arr,
                tau=0.75,  # some random value; doesn't affect prox outcome
            ),
            out=out,
        )

    @pytest.fixture
    def data_math_lipschitz(self):
        N_test, dim = 10, 2
        return self._random_array((N_test, dim))


class TestRangeSet(conftest.ProxFuncT):
    @pytest.fixture(  # test matrices
        params=[
            np.array(  # tall matrix
                [
                    [1, 0],
                    [0, 2],
                    [3, 0],
                    [0, 4],
                ]
            ),
            np.diag([1, 2, 3, 4]),  # square matrix
            np.array(  # fat matrix
                [
                    [1, 0, 3, 0],
                    [0, 2, 0, 4],
                ]
            ),
        ]
    )
    def A(self, request) -> np.ndarray:
        return request.param

    @pytest.fixture(
        params=itertools.product(
            pxd.NDArrayInfo,
            pxrt.Width,
        )
    )
    def spec(self, A, request) -> tuple[pxt.OpT, pxd.NDArrayInfo, pxrt.Width]:
        ndi, width = request.param
        self._skip_if_unsupported(ndi)
        xp = ndi.module()

        A = xp.array(A, dtype=width.value)
        A = pxa.LinOp.from_array(A, enable_warnings=False)

        op = pxo.RangeSet(A)
        return op, ndi, width

    @pytest.fixture
    def data_shape(self, A) -> pxt.OpShape:
        return (1, A.shape[0])

    @pytest.fixture(
        params=[  # (seed, in_set)
            (1, True),
            (1, False),
            (2, True),
            (2, False),
            (6, True),
            (6, False),
        ]
    )
    def data_apply(self, A, request):
        seed, in_set = request.param
        M, N = A.shape

        u, s, vh = np.linalg.svd(A)
        Q = u[(slice(None), *s.nonzero())]  # orth basis of range(A)
        Qp = np.eye(M) - Q @ Q.T  # orth basis of range(A)^\perp

        # override `in_set` if A is full rank.
        if (s > 0).sum() == min(M, N):
            in_set = True

        if in_set:
            # generate data point in range(A)
            arr = (Q @ Q.T) @ self._random_array((M,), seed=seed)
            out = np.r_[0]
        else:
            # generate data point in range(A)^\perp
            seek = True
            while seek:
                arr = Qp @ self._random_array((M,), seed=seed)
                seek = np.allclose(arr, 0)
            out = np.r_[np.inf]

        return dict(
            in_=dict(arr=arr),
            out=out,
        )

    @pytest.fixture(params=[0, 3, 7])  # seeds
    def data_prox(self, A, request) -> conftest.DataLike:
        M, _ = A.shape
        arr = self._random_array((M,), seed=request.param)

        Q, _ = np.linalg.qr(A)
        out = Q @ Q.T @ arr
        return dict(
            in_=dict(
                arr=arr,
                tau=1,  # some random value; doesn't affect prox outcome
            ),
            out=out,
        )

    @pytest.fixture
    def data_math_lipschitz(self, A):
        N_test, dim = 10, A.shape[0]
        return self._random_array((N_test, dim))


class TestAffineSet(conftest.ProxFuncT):
    @pytest.fixture(  # test matrices
        params=[
            np.diag([1, 2, 3, 4]),  # square matrix
            np.array(  # fat matrix
                [
                    [1, 0, 3, 0],
                    [0, 2, 0, 4],
                ]
            ),
        ]
    )
    def A(self, request) -> np.ndarray:
        return request.param

    @pytest.fixture(params=["zero", "span"])  # where b should lie
    def b(self, A, request) -> np.ndarray:
        M, N = A.shape
        if request.param == "zero":
            b = np.zeros((M,))
        else:  # span
            b = A @ self._random_array((N,))
        return b

    @pytest.fixture(
        params=itertools.product(
            pxd.NDArrayInfo,
            pxrt.Width,
        )
    )
    def spec(self, A, b, request) -> tuple[pxt.OpT, pxd.NDArrayInfo, pxrt.Width]:
        ndi, width = request.param
        self._skip_if_unsupported(ndi)
        xp = ndi.module()

        with pxrt.Precision(width):
            A = xp.array(A, dtype=width.value)
            A = pxa.LinOp.from_array(A, enable_warnings=False)
            b = xp.array(b, dtype=width.value)
            op = pxo.AffineSet(A, b)
        return op, ndi, width

    @pytest.fixture
    def data_shape(self, A) -> pxt.OpShape:
        return (1, A.shape[1])

    @pytest.fixture(
        params=[  # (seed, in_set)
            (1, True),
            (1, False),
            (6, True),
            (6, False),
        ]
    )
    def data_apply(self, A, b, request):
        seed, in_set = request.param
        M, N = A.shape

        u, s, vh = np.linalg.svd(A)

        # override `in_set` if A is square. -> precondition was to be full-rank
        if M == N:
            in_set = True

        if in_set:
            # generate data point s.t. Ax=b is true
            # seed is not used here; that's ok
            arr, *_ = np.linalg.lstsq(
                A.T @ A,
                A.T @ b,
                rcond=None,  # to silence NumPy's FutureWarning
            )
            out = np.r_[0]
        else:
            # generate data point in range(A)^\perp
            b = b + self._random_array((M,), seed=seed)
            arr, *_ = np.linalg.lstsq(
                A.T @ A,
                A.T @ b,
                rcond=None,  # to silence NumPy's FutureWarning
            )
            out = np.r_[np.inf]

        return dict(
            in_=dict(arr=arr),
            out=out,
        )

    @pytest.fixture(params=[0, 3, 7])  # seeds
    def data_prox(self, A, b, request) -> conftest.DataLike:
        M, N = A.shape
        arr = self._random_array((N,), seed=request.param)

        y, *_ = np.linalg.lstsq(
            A @ A.T,
            A @ arr - b,
            rcond=None,  # to silence NumPy's FutureWarning
        )
        out = arr - A.T @ y

        return dict(
            in_=dict(
                arr=arr,
                tau=1,  # some random value; doesn't affect prox outcome
            ),
            out=out,
        )

    @pytest.fixture
    def data_math_lipschitz(self, A):
        N_test, dim = 10, A.shape[1]
        return self._random_array((N_test, dim))



import collections.abc as cabc
import functools
import inspect

import dask
import dask.graph_manipulation as dgm

import pyxu.info.deps as pxd
import pyxu.info.ptype as pxt
import pyxu.util.array_module as pxa
import pyxu.util.misc as pxm

__all__ = [
    "as_canonical_shape",
    "vectorize",
]


def as_canonical_shape(x: pxt.NDArrayShape) -> pxt.NDArrayShape:
    """
    Transform a lone integer into a valid tuple-based shape specifier.
    """
    if isinstance(x, cabc.Iterable):
        x = tuple(x)
    else:
        x = (x,)
    sh = tuple(map(int, x))
    return sh


def vectorize(
    i: pxt.VarName,
    method: str = "scan",
    codim: pxt.Integer = None,
) -> cabc.Callable:
    """
    Decorator to auto-vectorize an array function to abide by :py:class:`~pyxu.abc.Property` API rules.

    Parameters
    ----------
    i: VarName
        Function parameter to vectorize. This variable must hold an object with a NumPy API.
    method: str
        Vectorization strategy:

        * `scan` computes outputs using a for-loop.
        * `parallel` passes inputs to DASK and evaluates them in parallel.
        * `scan_dask` passes inputs to DASK but evaluates inputs in sequence.
          This is useful if the function being vectorized has a shared resource, i.e. is not thread-safe.  It
          effectively gives a DASK-unaware function the ability to work with DASK inputs.
    codim: Integer
        Size of the function's core dimension output.

        This parameter is only required in "parallel" and "scan_dask" modes.

    Example
    -------
    .. code-block:: python3

       import pyxu.util as pxu

       @pxu.vectorize('x')
       def f(x):
           return x.sum(keepdims=True)

       x = np.arange(10).reshape((2, 5))
       f(x[0]), f(x[1])  #  [10], [35]
       f(x)              #  [10, 35] -> would have retured [45] if not decorated.
    """
    method = method.strip().lower()
    assert method in ("scan", "scan_dask", "parallel"), f"Unknown vectorization method '{method}'."
    if using_dask := (method in ("scan_dask", "parallel")):
        assert isinstance(codim, pxt.Integer), f"Parameter[codim] must be specified for DASK-backed '{method}'."

    def decorator(func: cabc.Callable) -> cabc.Callable:
        sig = inspect.Signature.from_callable(func)
        if i not in sig.parameters:
            error_msg = f"Parameter[{i}] not part of {func.__qualname__}() parameter list."
            raise ValueError(error_msg)

        @functools.wraps(func)
        def wrapper(*ARGS, **KWARGS):
            func_args = pxm.parse_params(func, *ARGS, **KWARGS)

            x = func_args.pop(i)
            *sh, dim = x.shape
            x = x.reshape((-1, dim))
            N, xp = len(x), pxa.get_array_module(x)

            if using_dask:
                f = lambda _: func(**{i: _, **func_args})
                blks = _dask_zip(
                    func=(f,) * N,
                    data=x,
                    out_shape=[(codim,)] * N,  # can't infer codim -> user-specified
                    out_dtype=(x.dtype,) * N,
                    parallel=method == "parallel",
                )
                y = xp.stack(blks, axis=0)  # (N, codim)
            else:  # method = "scan"
                # infer output dimensions + allocate
                func_args[i] = x[0]
                y0 = func(**func_args)
                y = xp.zeros((N, y0.size), dtype=y0.dtype)  # (N, codim)

                y[0] = y0
                for k in range(1, N):
                    func_args[i] = x[k]
                    y[k] = func(**func_args)

            y = y.reshape(*sh, -1)
            return y

        return wrapper

    return decorator


def _dask_zip(
    func: list[cabc.Callable],
    data: list[pxt.NDArray],
    out_shape: list[pxt.NDArrayShape],
    out_dtype: list[pxt.DType],
    parallel: bool,
) -> list[pxt.NDArray]:
    """
    Map functions in parallel via Dask.

    Computes the equivalent of ``out = [f(x) for (f, x) in zip(func, data)]``, with the following semantics:

    * If `data` contains only NUMPY/CUPY arrays, then ``out`` is computed as above.
    * If `data` contains only DASK arrays, then entries of ``out`` are computed:

      * if `parallel` enabled  -> dask-delay each `func`, then evaluate in parallel.
      * if `parallel` disabled -> dask-delay each `func`, then evaluate in sequence.
        (This is useful if `func` s share a common resource, thus not thread-safe.)

    For Dask-array inputs, this amounts to creating a task graph with virtual dependencies between successive `func`
    calls. In other words, the task graph looks like::

        _dask_zip(func, data, parallel=True) -> out

                      +----+              +----+
            data[0]-->|func|-->blk[0]-+-->|list|-->out
                .     +----+          |   +----+
                .                     |
                .     +----+          |
            data[n]-->|func|-->blk[n]-+
                      +----+

        ==========================================================================================================
        _dask_zip(func, data, parallel=False) -> out
                                                                                               +----+
                                +-------------------+----------------+--------------------+----+list|-->out
                                |                   |                |                    |    +----+
                                |                   |                |                    |
                      +----+    |        +----+     |      +---+     |         +----+     |
            data[0]-->|func|-->out[0]-+->|func|-->out[1]-->|...|-->out[n-1]-+->|func|-->out[n]
                      +----+          |  +----+            +---+            |  +----+
                                      |                                     |
            data[1]-------------------+                                     |
                .                                                           |
                .                                                           |
                .                                                           |
            data[n]---------------------------------------------------------+

    Parameters
    ----------
    func: list
        Functions to apply to each element of `data`.

        Function signatures are ``Callable[[NDArray], NDArray]``.
    data: list
        (N_data,) arrays to act on.
    out_shape: list
        Shapes of ``func[i](data[i])``.

        This parameter is only used if inputs are DASK arrays.  Its goal is to transform
        :py:class:`~dask.delayed.Delayed` objects back to array form.
    out_dtype: list
        Dtypes of ``func[i](data[i])``.

        This parameter is only used if inputs are DASK arrays.  Its goal is to transform
        :py:class:`~dask.delayed.Delayed` objects back to array form.

    Returns
    -------
    out: list
        (N_data,) objects acted upon.

        Outputs have the same backend/dtype as inputs, or as specified by `out_[shape,dtype]`.
    """
    assert all(len(_) == len(func) for _ in [data, out_shape, out_dtype])

    NDI = pxd.NDArrayInfo
    dask_input = lambda obj: NDI.from_obj(obj) == NDI.DASK
    if all(map(dask_input, data)):
        out = []
        for i in range(len(data)):
            if parallel:
                # If parallel=True, the user guarantees that functions CAN be executed in parallel,
                # i.e. no side-effects induced by calling the func[k] if arbitrary order.
                # The functions are hence PURE.
                _func = dask.delayed(func[i], pure=True)
            else:
                # If parallel=False, side-effects MAY influence func[k] outputs.
                # The functions are hence IMPURE.
                _func = dgm.bind(
                    children=dask.delayed(func[i], pure=False),
                    parents=out[i - 1] if (i > 0) else [],
                )
            blk = _array_ize(_func(data[i]), out_shape[i], out_dtype[i])
            out.append(blk)
    else:
        out = [_func(arr) for (_func, arr) in zip(func, data)]
    return out


def _array_ize(
    data,
    shape: pxt.NDArrayShape,
    dtype: pxt.DType,
) -> pxt.NDArray:
    """
    Transform a Dask-delayed object into its Dask-array counterpart.

    This function is a no-op if `data` is not a :py:class:`~dask.delayed.Delayed` object.

    Parameters
    ----------
    data: NDArray, :py:class:`~dask.delayed.Delayed`
        Object to act upon.
    shape: NDArrayShape
        Shape of `data`.

        This parameter is only used if `data` is a :py:class:`~dask.delayed.Delayed` object.  Its goal is to transform
        the former back to array form.
    dtype: DType
        Dtype of `data`.

        This parameter is only used if `data` is a :py:class:`~dask.delayed.Delayed` object.  Its goal is to transform
        the former back to array form.

    Returns
    -------
    arr: NDArray
        Dask-backed NDArray if `data` was a :py:class:`~dask.delayed.Delayed` object; no-op otherwise.
    """
    from dask.delayed import Delayed

    if isinstance(data, Delayed):
        xp = pxd.NDArrayInfo.DASK.module()
        arr = xp.from_delayed(
            data,
            shape=shape,
            dtype=dtype,
        )
    else:
        arr = data
    return arr



import itertools

import numpy as np
import pytest

import pyxu.abc as pxa
import pyxu.info.deps as pxd
import pyxu.operator as pxo
import pyxu.runtime as pxrt
import pyxu.util as pxu
import pyxu_tests.operator.conftest as conftest


class Sin(pxa.DiffMap):
    # f: \bR^{M} -> \bR^{M}
    #      x     -> sin(x)
    def __init__(self, M: int):
        super().__init__(shape=(M, M))
        self.lipschitz = 1
        self.diff_lipschitz = 1

    @pxrt.enforce_precision(i="arr")
    def apply(self, arr):
        xp = pxu.get_array_module(arr)
        y = xp.sin(arr)
        return y

    def jacobian(self, arr):
        xp = pxu.get_array_module(arr)
        J = xp.cos(arr)
        return pxo.DiagonalOp(J)


class TestSin(conftest.DiffMapT):
    @pytest.fixture(
        params=itertools.product(
            (  # dim, op
                (4, Sin(M=4)),
                (1, Sin(M=1)),
            ),
            pxd.NDArrayInfo,
            pxrt.Width,
        )
    )
    def _spec(self, request):
        return request.param

    @pytest.fixture
    def spec(self, _spec):
        return _spec[0][1], _spec[1], _spec[2]

    @pytest.fixture
    def dim(self, _spec):
        return _spec[0][0]

    @pytest.fixture
    def data_shape(self, dim):
        return (dim, dim)

    @pytest.fixture
    def data_apply(self, dim):
        A = np.linspace(0, np.pi / 2, dim)
        B = np.sin(A)
        return dict(
            in_=dict(arr=A),
            out=B,
        )

    @pytest.fixture
    def data_math_lipschitz(self, dim):
        N_test = 5
        return self._random_array((N_test, dim))

    @pytest.fixture
    def data_math_diff_lipschitz(self, dim):
        N_test = 5
        return self._random_array((N_test, dim))



import collections.abc as cabc
import datetime as dt
import functools
import itertools
import operator
import time

import numpy as np
import pytest
import scipy.optimize as sopt

import pyxu.abc as pxa
import pyxu.info.deps as pxd
import pyxu.info.ptype as pxt
import pyxu.opt.stop as pxst
import pyxu.runtime as pxrt
import pyxu.util as pxu
import pyxu_tests.conftest as ct


def funcs(N: int, seed: int = 0) -> cabc.Sequence[tuple[pxt.OpT, pxt.OpT]]:
    # Sequence of functional descriptors. (More terms can be added.)
    #
    # Used to create strongly-convex functionals.
    #
    # These functions MUST be backend-agnostic.
    import pyxu.operator as pxo
    import pyxu_tests.operator.examples.test_unitop as unitop

    L2 = pxo.SquaredL2Norm(dim=N)
    Id = pxo.IdentityOp(dim=N)

    rng = np.random.default_rng(seed)
    f = [
        (  # f1(x) = \norm{A1 x}{2}^{2}
            L2,
            pxo.HomothetyOp(cst=rng.uniform(1.1, 1.3), dim=N),
        ),
        (  # f2(x) = \norm{A2 x - y2}
            L2.asloss(rng.uniform(1, 3)),
            unitop.Permutation(N=N),
        ),
        (  # f3(x) = sum(x)
            pxo.Sum(arg_shape=N),
            Id,
        ),
        (  # f4(x) = \norm{a x}{2}^{2}
            L2.argscale(rng.uniform(-5, -1.1)),
            Id,
        ),
        (  # f5(x) = cst * \norm{x + y5}{2}^{2}
            L2.argshift(rng.normal()) * rng.uniform(1.1, 3),
            Id,
        ),
    ]
    return f


def generate_funcs(descr, N_term: int) -> cabc.Sequence[tuple[pxt.OpT]]:
    # Take description of many functionals, i.e. output of funcs(), and return a stream of
    # length-N_term tuples, where each term of the tuple is a functional created by summing a subset
    # of `descr`.
    #
    # Examples
    # --------
    # generate_funcs([(a, b), (c, d), (e, f)], 1)
    # -> [ (a*b + c*d + e*f,), ]
    #
    # generate_funcs([(a, b), (c, d), (e, f)], 2)
    # -> [ (a*b, c*d + e*f), (c*d, a*b + e*f), (e*f, a*b + c*d), ]
    assert 1 <= N_term <= len(descr)

    def chain(x, y):
        comp = x * y
        comp.diff_lipschitz = comp.estimate_diff_lipschitz(method="svd")
        return comp

    stream = []
    for d in itertools.permutations(descr, N_term - 1):
        # split into N_term subsets
        parts = [chain(*dd) for dd in d]  # The N_term - 1 first subsets each correspond to a single term
        # The remaining terms are summed in the last subset (no need to permute them)
        to_sum = list(set(descr) - set(d))
        p = functools.reduce(operator.add, [chain(*dd) for dd in to_sum])
        p.diff_lipschitz = p.estimate_diff_lipschitz(method="svd")
        parts.append(p)
        stream.append(tuple(parts))
    return stream


class SolverT(ct.DisableTestMixin):
    # Helper Functions --------------------------------------------------------
    @staticmethod
    def _check_allclose(nd_1: dict, nd_2: dict) -> bool:
        same_keys = set(nd_1.keys()) == set(nd_2.keys())
        if not same_keys:
            return False

        def standardize(x: pxt.NDArray) -> np.ndarray:
            # _check_allclose() compares inputs for equality using np.allclose().
            # Problem: we want NaNs to compare equal, which is not the case with np.allclose().
            # [Whether this makes sense is test-dependant: safeguards are placed where relevant.]
            # We therefore use np.nan_to_num() to bound potential NaNs and make them compare
            # equally.
            y = pxu.to_NUMPY(x)
            z = np.nan_to_num(y)
            return z

        stats = dict()
        for k in nd_1.keys():
            stats[k] = np.allclose(
                standardize(nd_1[k]),
                standardize(nd_2[k]),
            )
        return all(stats.values())

    @staticmethod
    def _check_prec_match(nd_1: dict, nd_2: dict) -> bool:
        same_keys = set(nd_1.keys()) == set(nd_2.keys())
        if not same_keys:
            return False

        stats = dict()
        for k in nd_1.keys():
            stats[k] = nd_1[k].dtype == nd_2[k].dtype
        return all(stats.values())

    @staticmethod
    def as_early_stop(kwargs: dict) -> dict:
        # Some tests look at state which does not require a solver to have converged
        # (mathematically).
        # This function adds a max-iter constraint to the kwargs_fit() dictionary to drastically
        # curb test time.
        kwargs = kwargs.copy()
        c1 = pxst.MaxIter(n=5)
        c2 = pxst.MaxDuration(t=dt.timedelta(seconds=5))
        kwargs["stop_crit"] = c1 | c2
        return kwargs

    # Fixtures ----------------------------------------------------------------
    @pytest.fixture
    def spec(self) -> tuple[pxt.SolverC, dict, dict]:
        # override in subclass to return:
        # * the operator (class) to test;
        # * the solver-specific kwargs to pass to __init__().
        # * the solver-specific kwargs to pass to fit().
        #
        # The triplet (solver_klass, kwargs_init, kwargs_fit) must be provided since we need to
        # manipulate solver creation/execution.
        #
        # IMPORTANT: it is assumed operators involved in the solver's instantiation are
        # backend-agnostic.
        raise NotImplementedError

    @pytest.fixture
    def solver_klass(self, spec) -> pxt.SolverC:
        return spec[0]

    @pytest.fixture
    def kwargs_init(self, spec) -> dict:
        kwargs = spec[1]
        kwargs.update(show_progress=False)
        return kwargs

    @pytest.fixture
    def kwargs_fit(self, spec) -> dict:
        return spec[2]

    @pytest.fixture
    def _kwargs_fit_xp(self, kwargs_fit, xp) -> dict:
        # Same as `kwargs_fit`, but with all NDArray inputs transformed to backend[xp].
        # For internal use only: do not override in sub-classes.
        data = kwargs_fit.copy()
        for k in data.keys():
            v = data[k]
            try:
                pxu.get_array_module(v, fallback=None)
                data[k] = xp.array(v)
            except Exception:
                # Not an NDArray -> no transformation
                pass
        return data

    @pytest.fixture
    def solver(self, solver_klass, kwargs_init) -> pxt.SolverT:
        # Solver instance used for most fit() tests.
        slvr = solver_klass(**kwargs_init)
        return slvr

    @pytest.fixture(
        # We do not run solvers with DASK inputs due to general slowness.
        # This is not a problem in practice: it is mostly operators which
        # need to be tested for DASK compliance.
        params=[
            pxd.NDArrayInfo.NUMPY.module(),
            pytest.param(
                pxd.NDArrayInfo.CUPY.module(),
                marks=pytest.mark.skipif(
                    not pxd.CUPY_ENABLED,
                    reason="GPU unsupported on this machine.",
                ),
            ),
        ]
    )
    def xp(self, request) -> pxt.ArrayModule:
        return request.param

    @pytest.fixture(params=pxrt.Width)
    def width(self, request) -> pxrt.Width:
        return request.param

    @pytest.fixture
    def cost_function(self, kwargs_init) -> dict[str, pxt.OpT]:
        # override in subclass to create cost function minimized per logged variable.
        #
        # Inputs are assumed to be some linear combination of conftest.funcs().
        # The cost functions are hence guaranteed to be a strongly convex.
        raise NotImplementedError

    @pytest.fixture
    def ground_truth(self, xp, cost_function) -> dict:
        # Compute the optimal value of the cost functions.
        #
        # Must return the same output as solver.stats()[0], i.e. the data dictionary.
        def fun(x: np.ndarray, cost_f: pxt.OpT) -> tuple[float, np.ndarray]:  # f(x), \grad_{f}(x)
            val = cost_f.apply(x)
            grad = cost_f.grad(x)
            return float(val), grad

        rng = np.random.default_rng()
        data = dict()
        for log_var, cost_f in cost_function.items():
            res = sopt.minimize(
                fun=fun,
                args=(cost_f,),
                x0=rng.uniform(size=cost_f.dim),
                method="CG",
                jac=True,
            )
            assert res.success  # if raised, means cost function was not strongly-convex
            data[log_var] = xp.array(res.x)
        return data

    # Tests -------------------------------------------------------------------
    def test_backend_fit(self, solver, _kwargs_fit_xp, xp):
        # solver output-backend match inputs
        self._skip_if_disabled()

        solver.fit(**self.as_early_stop(_kwargs_fit_xp))
        data, _ = solver.stats()

        stats = {k: pxu.get_array_module(v) == xp for (k, v) in data.items()}
        assert all(stats.values())

    def test_precCM_fit(self, solver, kwargs_fit, width):
        # solver output-precision match context manager
        self._skip_if_disabled()

        with pxrt.Precision(width):
            solver.fit(**self.as_early_stop(kwargs_fit))
            data, _ = solver.stats()

            stats = {k: v.dtype == width.value for (k, v) in data.items()}
            assert all(stats.values())

    @pytest.mark.parametrize("eps_threshold", [5e-3])
    @pytest.mark.parametrize("time_threshold", [dt.timedelta(seconds=5)])
    def test_value_fit(
        self,
        solver,
        _kwargs_fit_xp,
        width,
        cost_function,
        ground_truth,
        eps_threshold,  # rel-threshold
        time_threshold,  # max runtime
    ):
        # Ensure algorithm converges to ground-truth. (All backends/precisions.)
        self._skip_if_disabled()

        # SciPy ground truth may have converged closer to the solution than `solver` due to
        # different stopping criteria used. (Default or user-specified.)
        #
        # Since cost-functions are assumed strongly convex, a solver is assumed to converge to the
        # ground-truth if the following conditions hold:
        #
        #   1. |x_k+1 - x_k| <= eps |x_k|
        #   2. f(x_gt) <= f(x_N) < f(x_0), where `f` denotes the cost function
        #   3. max runtime <= threshold
        #
        # (1) ensures solver iterates form a Cauchy sequence, hence converge to some limit.
        # (2) ensures the cost function decreases, hence the solver is converging to something.
        # (3) just avoids a solver running forever.
        #
        # NOTE: f(x_N) < f(x_0) implies initial_points specified in kwargs_fit() should be chosen
        # `far` from the ground-truth.
        # Test writers are responsible of setting x_0 such that this condition holds.
        var_crit = [
            pxst.RelError(
                eps=eps_threshold,
                var=k,
                f=None,
                satisfy_all=True,
            )
            for k in cost_function.keys()
        ]
        time_crit = pxst.MaxDuration(t=time_threshold)

        kwargs_fit = _kwargs_fit_xp.copy()
        kwargs_fit.update(
            stop_crit=time_crit | functools.reduce(operator.and_, var_crit),
            mode=pxa.SolverMode.MANUAL,
        )
        with pxrt.Precision(width):
            solver.fit(**kwargs_fit)

        data_0, _ = solver.stats()
        for data in solver.steps():
            pass
        data_opt, history = solver.stats()

        # ensure solver ended because rel-err threshold reached, not timeout
        relerr_decrease = dict()
        for var in cost_function.keys():
            if data_0[var].ndim == 1:
                k = f"RelError[{var}]"
            else:
                k = f"RelError[{var}]_max"
            relerr_decrease[var] = history[k][-1] <= eps_threshold
        assert all(relerr_decrease.values())

        # ensure cost-function decreased from start to end
        cost_decrease = dict()
        for var, c_func in cost_function.items():
            c_0 = c_func.apply(data_0[var])
            c_opt = c_func.apply(data_opt[var])
            c_gt = c_func.apply(ground_truth[var])

            # Numerical errors can make c_gt > c_opt at convergence.
            # We piggy-back on less_equal() from the operator test suite to handle this case.
            bound_1 = ct.less_equal(c_gt, c_opt, as_dtype=width.value)

            # The cost function MUST decrease. (See test-level comment for details.)
            bound_2 = c_opt < c_0

            cost_decrease[var] = np.all(bound_1 & bound_2)
        assert all(cost_decrease.values())

    def test_transparent_fit(self, solver, kwargs_fit):
        # Running solver twice returns same results.
        self._skip_if_disabled()

        # `kwargs_fit` values may be coerced on 1st call to solver.fit(), in which case transparency
        # will not be tested. (May happen if solver takes `kwargs_fit` values as-is, and does not
        # make a copy internally.)
        # Solution: Manually enforce precision of all input arrays prior to calling solver.fit().
        kw_fit = dict()
        for k, v in kwargs_fit.items():
            try:
                v_c = pxrt.coerce(v)
            except Exception:
                v_c = v
            kw_fit[k] = v_c
        kw_fit = self.as_early_stop(kw_fit)

        solver.fit(**kw_fit.copy())
        data1, _ = solver.stats()
        solver.fit(**kw_fit.copy())
        data2, _ = solver.stats()

        assert self._check_allclose(data1, data2)

    # -------------------------------------------------------------------------
    # The Solver API has a complex structure where sub-classes are expected to override a subset of
    # methods, i.e m_init(), m_step(), default_stop_crit(), objective_func(). The tests below ensure
    # Solver sub-classes do not break the Solver API by overriding forbidden methods.

    @pytest.mark.parametrize("track_objective", [True, False])
    def test_objective_func_tracked(
        self,
        solver,
        kwargs_fit,
        track_objective,
    ):
        # Ensure objective_func value present in history.
        self._skip_if_disabled()

        kwargs_fit = self.as_early_stop(kwargs_fit)
        kwargs_fit.update(track_objective=track_objective)
        solver.fit(**kwargs_fit)
        _, history = solver.stats()

        tracked = any("objective_func" in f for f in history.dtype.fields)
        if track_objective:
            assert tracked
        else:
            assert not tracked

    @pytest.mark.parametrize("stop_rate", [1, 2, 3])
    def test_stop_rate_limits_history(
        self,
        solver_klass,
        kwargs_init,
        kwargs_fit,
        stop_rate,
    ):
        # stop_rate affects history granularity.
        self._skip_if_disabled()

        kwargs_init = kwargs_init.copy()
        kwargs_init.update(stop_rate=stop_rate)
        solver = solver_klass(**kwargs_init)

        kwargs_fit = kwargs_fit.copy()
        kwargs_fit["stop_crit"] = pxst.MaxIter(n=10 * stop_rate)  # avoids infinite loop if solver does not converge.
        solver.fit(**kwargs_fit)

        _, history = solver.stats()
        assert np.all(history["iteration"] % stop_rate == 0)

    def test_data_contains_logvar(self, solver, kwargs_fit):
        # logged data only contains variables from log_var.
        self._skip_if_disabled()

        log_var = solver._astate["log_var"]
        solver.fit(**self.as_early_stop(kwargs_fit))
        data, _ = solver.stats()
        assert set(log_var) == set(data.keys())

    def test_halt_implies_disk_storage(self, solver_klass, kwargs_init, kwargs_fit, tmp_path):
        # When solver stops, data+log files exist at specified folder.
        self._skip_if_disabled()

        kwargs_init = kwargs_init.copy()
        kwargs_init.update(folder=tmp_path, exist_ok=True)
        solver = solver_klass(**kwargs_init)
        solver.fit(**self.as_early_stop(kwargs_fit))

        assert solver.workdir.resolve() == tmp_path.resolve()
        assert solver.logfile.exists()
        assert solver.datafile.exists()

    def test_disk_value_matches_memory(self, solver, kwargs_fit):
        # Datafile content (values) match in-memory data after halt.
        self._skip_if_disabled()

        solver.fit(**self.as_early_stop(kwargs_fit))

        disk = np.load(solver.datafile)
        data_disk = {k: v for (k, v) in disk.items() if k != "history"}
        hist_disk = disk["history"]
        data_mem, hist_mem = solver.stats()

        assert self._check_allclose(data_disk, data_mem)
        assert self._check_allclose(
            # transform structured arrays to dict.
            {k: hist_disk[k] for k in hist_disk.dtype.fields},
            {k: hist_mem[k] for k in hist_mem.dtype.fields},
        )

    def test_disk_prec_matches_memory(self, solver, kwargs_fit, width):
        # Datafile content (dtypes) match in-memory dtypes after halt.
        self._skip_if_disabled()

        with pxrt.Precision(width):
            solver.fit(**self.as_early_stop(kwargs_fit))

            disk = np.load(solver.datafile)
            data_disk = {k: v for (k, v) in disk.items() if k != "history"}
            data_mem, _ = solver.stats()

            assert self._check_prec_match(data_disk, data_mem)

    def test_transparent_mode(self, solver, kwargs_fit):
        # All execution modes return same results.
        self._skip_if_disabled()

        data = dict()
        for m in [pxa.SolverMode.BLOCK, pxa.SolverMode.MANUAL]:
            kwargs_fit = self.as_early_stop(kwargs_fit)
            kwargs_fit.update(mode=m)
            solver.fit(**kwargs_fit)
            if m == pxa.SolverMode.BLOCK:
                pass
            elif m == pxa.SolverMode.ASYNC:
                while solver.busy():
                    time.sleep(0.5)
                solver.stop()
            else:  # m == pxa.SolverMode.MANUAL
                for _ in solver.steps():
                    pass
            d, _ = solver.stats()
            data[m] = d

        stats = dict()
        for k1, k2 in itertools.combinations(data.keys(), r=2):
            stats[(k1, k2)] = self._check_allclose(data[k1], data[k2])
        assert all(stats.values())



# How AddRule tests work:
#
# * AddRuleMixin auto-defines all arithmetic method (input,output) pairs.
#   [Caveat: we assume all tested examples are defined on \bR.] (This is not a problem in practice.)
#   [Caveat: we assume the base operators (op_lhs, op_rhs) are correctly implemented.] (True if choosing test operators from examples/.)
#
# * To test a compound operator (via +), inherit from AddRuleMixin and the suitable conftest.MapT
#   subclass which the compound operator should abide by.


import collections.abc as cabc
import itertools

import numpy as np
import pytest
import scipy.linalg as splinalg

import pyxu.abc as pxa
import pyxu.info.deps as pxd
import pyxu.info.ptype as pxt
import pyxu.runtime as pxrt
import pyxu.util as pxu
import pyxu_tests.operator.conftest as conftest


# LHS/RHS test operators ------------------------------------------------------
def op_map():
    import pyxu_tests.operator.examples.test_map as tc

    return tc.ReLU(M=7)


def op_func():
    import pyxu_tests.operator.examples.test_func as tc

    return tc.Median(dim=7)


def op_diffmap():
    import pyxu_tests.operator.examples.test_diffmap as tc

    return tc.Sin(M=7)


def op_difffunc():
    import pyxu_tests.operator.examples.test_difffunc as tc

    return tc.SquaredL2Norm(M=7)


def op_proxfunc():
    import pyxu_tests.operator.examples.test_proxfunc as tc

    return tc.L1Norm(M=7)


def op_proxdifffunc():
    import pyxu_tests.operator.examples.test_proxdifffunc as tc

    return tc.SquaredL2Norm(M=7)


def op_quadraticfunc():
    from pyxu_tests.operator.examples.test_linfunc import ScaledSum
    from pyxu_tests.operator.examples.test_posdefop import PSDConvolution

    N = 7
    return pxa.QuadraticFunc(
        shape=(1, N),
        Q=PSDConvolution(N=N),
        c=ScaledSum(N=N),
        t=1,
    )


def op_linop(dim: int = 7, codim_scale: int = 1):
    import pyxu_tests.operator.examples.test_linop as tc

    return tc.Tile(N=dim, M=codim_scale)


def op_linfunc():
    import pyxu_tests.operator.examples.test_linfunc as tc

    return tc.ScaledSum(N=7)


def op_squareop():
    import pyxu_tests.operator.examples.test_squareop as tc

    return tc.CumSum(N=7)


def op_normalop():
    import pyxu_tests.operator.examples.test_normalop as tc

    rng = np.random.default_rng(seed=2)
    h = rng.normal(size=(7,))
    return tc.CircularConvolution(h=h)


def op_unitop():
    import pyxu_tests.operator.examples.test_unitop as tc

    return tc.Permutation(N=7)


def op_selfadjointop():
    import pyxu_tests.operator.examples.test_selfadjointop as tc

    return tc.SelfAdjointConvolution(N=7)


def op_posdefop():
    import pyxu_tests.operator.examples.test_posdefop as tc

    return tc.PSDConvolution(N=7)


def op_projop():
    import pyxu_tests.operator.examples.test_projop as tc

    return tc.Oblique(N=7, alpha=0.1)


def op_orthprojop():
    import pyxu_tests.operator.examples.test_orthprojop as tc

    return tc.ScaleDown(N=7)


# Data Mixin ------------------------------------------------------------------
class AddRuleMixin:
    # Fixtures ----------------------------------------------------------------
    @pytest.fixture
    def op_lrhs(self) -> tuple[pxt.OpT, pxt.OpT]:
        # Override in inherited class with LHS/RHS operands.
        raise NotImplementedError

    @pytest.fixture
    def op_lhs(self, op_lrhs) -> pxt.OpT:
        return op_lrhs[0]

    @pytest.fixture
    def op_rhs(self, op_lrhs) -> pxt.OpT:
        return op_lrhs[1]

    @pytest.fixture(
        params=itertools.product(
            [
                "lhs + rhs",
                "rhs + lhs",
            ],
            pxd.NDArrayInfo,
            pxrt.Width,
        )
    )
    def spec(self, op_lhs, op_rhs, request) -> tuple[pxt.OpT, pxd.NDArrayInfo, pxrt.Width]:
        mode, ndi, width = request.param
        if mode == "lhs + rhs":
            op = op_lhs + op_rhs
        elif mode == "rhs + lhs":
            op = op_rhs + op_lhs
        else:
            raise NotImplementedError
        return op, ndi, width

    @pytest.fixture
    def data_shape(self, op_lhs, op_rhs) -> pxt.OpShape:
        sh = pxu.infer_sum_shape(op_lhs.shape, op_rhs.shape)
        return sh

    @pytest.fixture
    def data_apply(self, op, op_lhs, op_rhs) -> conftest.DataLike:
        dim = self._sanitize(op.dim, 7)
        arr = self._random_array((dim,), seed=20)  # random seed for reproducibility
        out = op_lhs.apply(arr) + op_rhs.apply(arr)
        return dict(
            in_=dict(arr=arr),
            out=out,
        )

    @pytest.fixture
    def data_adjoint(self, op, op_lhs, op_rhs) -> conftest.DataLike:
        arr = self._random_array((op.codim,), seed=20)  # random seed for reproducibility
        if op_lhs.has(pxa.Property.FUNCTIONAL) and (not op_rhs.has(pxa.Property.FUNCTIONAL)):
            # LHS broadcasts
            out_lhs = op_lhs.adjoint(arr.sum(keepdims=True))
            out_rhs = op_rhs.adjoint(arr)
        elif (not op_lhs.has(pxa.Property.FUNCTIONAL)) and op_rhs.has(pxa.Property.FUNCTIONAL):
            # RHS broadcasts
            out_lhs = op_lhs.adjoint(arr)
            out_rhs = op_rhs.adjoint(arr.sum(keepdims=True))
        else:
            out_lhs = op_lhs.adjoint(arr)
            out_rhs = op_rhs.adjoint(arr)

        out = out_lhs + out_rhs
        return dict(
            in_=dict(arr=arr),
            out=out,
        )

    @pytest.fixture
    def data_grad(self, op, op_lhs, op_rhs) -> conftest.DataLike:
        dim = self._sanitize(op.dim, 7)
        arr = self._random_array((dim,), seed=20)  # random seed for reproducibility
        out = op_lhs.grad(arr) + op_rhs.grad(arr)
        return dict(
            in_=dict(arr=arr),
            out=out,
        )

    @pytest.fixture
    def data_prox(self, op, op_lhs, op_rhs) -> conftest.DataLike:
        dim = self._sanitize(op.dim, 7)
        arr = self._random_array((dim,), seed=20)  # random seed for reproducibility
        tau = np.abs(self._random_array((1,), seed=21))[0]  # random seed for reproducibility

        if op.has(pxa.Property.PROXIMABLE):
            out = None
            if op_lhs.has(pxa.Property.LINEAR):
                P, G = op_rhs, op_lhs
                out = P.prox(arr - tau * G.grad(arr), tau)
            elif op_rhs.has(pxa.Property.LINEAR):
                P, G = op_lhs, op_rhs
                out = P.prox(arr - tau * G.grad(arr), tau)
            elif op_lhs.has(pxa.Property.QUADRATIC) and op_rhs.has(pxa.Property.QUADRATIC):
                A = op_lhs._quad_spec()[0].asarray()
                B = op_rhs._quad_spec()[0].asarray()
                C = op_lhs.grad(np.zeros(op.dim))
                D = op_rhs.grad(np.zeros(op.dim))
                Q = tau * (A + B) + np.eye(op.dim)
                b = arr - tau * (C + D)
                out, *_ = splinalg.lstsq(Q, b)

            if out is not None:
                return dict(
                    in_=dict(
                        arr=arr,
                        tau=tau,
                    ),
                    out=out,
                )
        raise NotImplementedError

    @pytest.fixture
    def data_math_lipschitz(self, op) -> cabc.Collection[np.ndarray]:
        N_test, dim = 5, self._sanitize(op.dim, 7)
        return self._random_array((N_test, dim))

    @pytest.fixture
    def data_math_diff_lipschitz(self, op) -> cabc.Collection[np.ndarray]:
        N_test, dim = 5, self._sanitize(op.dim, 7)
        return self._random_array((N_test, dim))

    # Tests -------------------------------------------------------------------
    def test_interface_asloss(self, op, xp, width, op_lhs, op_rhs):
        self._skip_if_disabled()
        if pxa.Property.FUNCTIONAL not in (op_lhs.properties() & op_rhs.properties()):
            pytest.skip("asloss() unavailable for non-functionals.")

        try:
            op_lhs.asloss()  # detect if fails
            op_rhs.asloss()  # detect if fails
            super().test_interface_asloss(op, xp, width)
        except NotImplementedError:
            pytest.skip("asloss() unsupported by LHS/RHS operator(s).")


# Test classes (Maps) ---------------------------------------------------------
class TestAddRuleMap(AddRuleMixin, conftest.MapT):
    @pytest.fixture(
        params=[
            (op_map(), op_map()),
            (op_map(), op_func()),
            (op_map(), op_diffmap()),
            (op_map(), op_difffunc()),
            (op_map(), op_proxfunc()),
            (op_map(), op_proxdifffunc()),
            (op_map(), op_quadraticfunc()),
            (op_map(), op_linop()),
            (op_map(), op_linfunc()),
            (op_map(), op_squareop()),
            (op_map(), op_normalop()),
            (op_map(), op_unitop()),
            (op_map(), op_selfadjointop()),
            (op_map(), op_posdefop()),
            (op_map(), op_projop()),
            (op_map(), op_orthprojop()),
            (op_func(), op_diffmap()),
            (op_func(), op_linop()),
            (op_func(), op_squareop()),
            (op_func(), op_normalop()),
            (op_func(), op_unitop()),
            (op_func(), op_selfadjointop()),
            (op_func(), op_posdefop()),
            (op_func(), op_projop()),
            (op_func(), op_orthprojop()),
            (op_diffmap(), op_proxfunc()),
            (op_proxfunc(), op_linop()),
            (op_proxfunc(), op_squareop()),
            (op_proxfunc(), op_normalop()),
            (op_proxfunc(), op_unitop()),
            (op_proxfunc(), op_selfadjointop()),
            (op_proxfunc(), op_posdefop()),
            (op_proxfunc(), op_projop()),
            (op_proxfunc(), op_orthprojop()),
        ]
    )
    def op_lrhs(self, request):
        return request.param


class TestAddRuleDiffMap(AddRuleMixin, conftest.DiffMapT):
    @pytest.fixture(
        params=[
            (op_diffmap(), op_diffmap()),
            (op_diffmap(), op_difffunc()),
            (op_diffmap(), op_proxdifffunc()),
            (op_diffmap(), op_quadraticfunc()),
            (op_diffmap(), op_linop()),
            (op_diffmap(), op_linfunc()),
            (op_diffmap(), op_squareop()),
            (op_diffmap(), op_normalop()),
            (op_diffmap(), op_unitop()),
            (op_diffmap(), op_selfadjointop()),
            (op_diffmap(), op_posdefop()),
            (op_diffmap(), op_projop()),
            (op_diffmap(), op_orthprojop()),
            (op_difffunc(), op_linop()),
            (op_difffunc(), op_squareop()),
            (op_difffunc(), op_normalop()),
            (op_difffunc(), op_unitop()),
            (op_difffunc(), op_selfadjointop()),
            (op_difffunc(), op_posdefop()),
            (op_difffunc(), op_projop()),
            (op_difffunc(), op_orthprojop()),
            (op_quadraticfunc(), op_linop()),
            (op_quadraticfunc(), op_squareop()),
            (op_quadraticfunc(), op_normalop()),
            (op_quadraticfunc(), op_unitop()),
            (op_quadraticfunc(), op_selfadjointop()),
            (op_quadraticfunc(), op_posdefop()),
            (op_quadraticfunc(), op_projop()),
            (op_quadraticfunc(), op_orthprojop()),
            (op_quadraticfunc(), op_linop()),
            (op_quadraticfunc(), op_squareop()),
            (op_quadraticfunc(), op_normalop()),
            (op_quadraticfunc(), op_unitop()),
            (op_quadraticfunc(), op_selfadjointop()),
            (op_quadraticfunc(), op_posdefop()),
            (op_quadraticfunc(), op_projop()),
            (op_quadraticfunc(), op_orthprojop()),
        ]
    )
    def op_lrhs(self, request):
        return request.param


class TestAddRuleLinOp(AddRuleMixin, conftest.LinOpT):
    @pytest.fixture(
        params=[
            (op_linop(), op_linop()),
            (op_linop(), op_linfunc()),
            (op_linop(codim_scale=2), op_linop(codim_scale=2)),
            (op_linop(codim_scale=2), op_linfunc()),
        ]
    )
    def op_lrhs(self, request):
        return request.param


class TestAddRuleSquareOp(AddRuleMixin, conftest.SquareOpT):
    @pytest.fixture(
        params=[
            (op_linfunc(), op_squareop()),
            (op_linfunc(), op_normalop()),
            (op_linfunc(), op_unitop()),
            (op_linfunc(), op_selfadjointop()),
            (op_linfunc(), op_posdefop()),
            (op_linfunc(), op_projop()),
            (op_linfunc(), op_orthprojop()),
            (op_squareop(), op_squareop()),
            (op_squareop(), op_normalop()),
            (op_squareop(), op_unitop()),
            (op_squareop(), op_selfadjointop()),
            (op_squareop(), op_posdefop()),
            (op_squareop(), op_projop()),
            (op_squareop(), op_orthprojop()),
            (op_normalop(), op_normalop()),
            (op_normalop(), op_unitop()),
            (op_normalop(), op_selfadjointop()),
            (op_normalop(), op_posdefop()),
            (op_normalop(), op_projop()),
            (op_normalop(), op_orthprojop()),
            (op_unitop(), op_unitop()),
            (op_unitop(), op_selfadjointop()),
            (op_unitop(), op_posdefop()),
            (op_unitop(), op_projop()),
            (op_unitop(), op_orthprojop()),
            (op_selfadjointop(), op_projop()),
            (op_posdefop(), op_projop()),
            (op_projop(), op_projop()),
            (op_projop(), op_orthprojop()),
        ]
    )
    def op_lrhs(self, request):
        return request.param


class TestAddRuleNormalOp(AddRuleMixin, conftest.NormalOpT):
    @pytest.fixture(params=[])
    def op_lrhs(self, request):
        return request.param


class TestAddRuleUnitOp(AddRuleMixin, conftest.UnitOpT):
    @pytest.fixture(params=[])
    def op_lrhs(self, request):
        return request.param


class TestAddRuleSelfAdjointOp(AddRuleMixin, conftest.SelfAdjointOpT):
    @pytest.fixture(
        params=[
            (op_selfadjointop(), op_selfadjointop()),
            (op_selfadjointop(), op_posdefop()),
            (op_selfadjointop(), op_orthprojop()),
            (op_orthprojop(), op_orthprojop()),
        ]
    )
    def op_lrhs(self, request):
        return request.param


class TestAddRulePosDefOp(AddRuleMixin, conftest.PosDefOpT):
    @pytest.fixture(
        params=[
            (op_posdefop(), op_posdefop()),
            (op_posdefop(), op_orthprojop()),
        ]
    )
    def op_lrhs(self, request):
        return request.param


class TestAddRuleProjOp(AddRuleMixin, conftest.ProjOpT):
    @pytest.fixture(params=[])
    def op_lrhs(self, request):
        return request.param


class TestAddRuleOrthProjOp(AddRuleMixin, conftest.OrthProjOpT):
    @pytest.fixture(params=[])
    def op_lrhs(self, request):
        return request.param


# Test classes (Funcs) --------------------------------------------------------
class TestAddRuleFunc(AddRuleMixin, conftest.FuncT):
    @pytest.fixture(
        params=[
            (op_func(), op_func()),
            (op_func(), op_difffunc()),
            (op_func(), op_proxfunc()),
            (op_func(), op_proxdifffunc()),
            (op_func(), op_quadraticfunc()),
            (op_func(), op_linfunc()),
            (op_difffunc(), op_proxfunc()),
            (op_proxfunc(), op_proxfunc()),
            (op_proxfunc(), op_proxdifffunc()),
            (op_proxfunc(), op_quadraticfunc()),
        ]
    )
    def op_lrhs(self, request):
        return request.param


class TestAddRuleDiffFunc(AddRuleMixin, conftest.DiffFuncT):
    @pytest.fixture(
        params=[
            (op_difffunc(), op_difffunc()),
            (op_difffunc(), op_proxdifffunc()),
            (op_difffunc(), op_quadraticfunc()),
            (op_difffunc(), op_linfunc()),
            (op_proxdifffunc(), op_proxdifffunc()),
            (op_proxdifffunc(), op_quadraticfunc()),
        ]
    )
    def op_lrhs(self, request):
        return request.param


class TestAddRuleProxFunc(AddRuleMixin, conftest.ProxFuncT):
    @pytest.fixture(
        params=[
            (op_proxfunc(), op_linfunc()),
        ]
    )
    def op_lrhs(self, request):
        return request.param


class TestAddRuleProxDiffFunc(AddRuleMixin, conftest.ProxDiffFuncT):
    @pytest.fixture(
        params=[
            (op_proxdifffunc(), op_linfunc()),
        ]
    )
    def op_lrhs(self, request):
        return request.param


class TestAddRuleQuadraticFunc(AddRuleMixin, conftest.QuadraticFuncT):
    @pytest.fixture(
        params=[
            (op_quadraticfunc(), op_quadraticfunc()),
            (op_quadraticfunc(), op_linfunc()),
        ]
    )
    def op_lrhs(self, request):
        return request.param


class TestAddRuleLinFunc(AddRuleMixin, conftest.LinFuncT):
    @pytest.fixture(
        params=[
            (op_linfunc(), op_linfunc()),
        ]
    )
    def op_lrhs(self, request):
        return request.param



import pyxu.abc as pxa
import pyxu.info.ptype as pxt
import pyxu.runtime as pxrt
import pyxu.util as pxu

__all__ = [
    "backtracking_linesearch",
]


LINESEARCH_DEFAULT_R = 0.5
LINESEARCH_DEFAULT_C = 1e-4


@pxrt.enforce_precision(
    i=("x", "direction", "gradient", "a0", "r", "c"),
    allow_None=True,
)
def backtracking_linesearch(
    f: pxa.DiffFunc,
    x: pxt.NDArray,
    direction: pxt.NDArray,
    gradient: pxt.NDArray = None,
    a0: pxt.Real = None,
    r: pxt.Real = LINESEARCH_DEFAULT_R,
    c: pxt.Real = LINESEARCH_DEFAULT_C,
) -> pxt.NDArray:
    r"""
    Backtracking line search algorithm based on the `Armijo-Goldstein condition
    <https://www.wikiwand.com/en/Backtracking_line_search>`_.

    Parameters
    ----------
    f: ~pyxu.abc.operator.DiffFunc
        Differentiable functional.
    x: NDArray
        (..., N) initial search point(s).
    direction: NDArray
        (..., N) search direction(s) corresponding to initial point(s).
    gradient: NDArray
        (..., N) gradient of `f` at initial search point(s).

        Specifying `gradient` when known is an optimization: it will be autocomputed via
        :py:meth:`~pyxu.abc.DiffFunc.grad` if unspecified.
    a0: Real
        Initial step size.

        If unspecified and :math:`\nabla f` is :math:`\beta`-Lipschitz continuous, then `a0` is auto-chosen as
        :math:`\frac{1}{\beta}`.
    r: Real
        Step reduction factor.
    c: Real
        Bound reduction factor.

    Returns
    -------
    a: NDArray
        (..., 1) step sizes.
    """
    assert 0 < r < 1
    assert 0 < c < 1
    if a0 is None:
        a0 = pxrt.coerce(1 / f.diff_lipschitz)
        assert a0 > 0, "a0: cannot auto-set step size."
    else:
        assert a0 > 0

    if gradient is None:
        gradient = f.grad(x)

    f_x = f.apply(x)
    d_f = c * (gradient * direction).sum(axis=-1, keepdims=True)  # \delta f

    def refine(a: pxt.NDArray) -> pxt.NDArray:
        # Do one iteration of the algorithm.
        #
        # Parameters
        # ----------
        # a : NDArray
        #     (..., 1) current step size(s)
        #
        # Returns
        # -------
        # mask : NDArray[bool]
        #     (..., 1) refinement points
        lhs = f.apply(x + a * direction)
        rhs = f_x + a * d_f
        return lhs > rhs  # mask

    xp = pxu.get_array_module(x)
    a = xp.full(shape=(*x.shape[:-1], 1), fill_value=a0, dtype=x.dtype)
    while (mask := refine(a)).any():
        a[mask] *= r
    return a



import collections
import collections.abc as cabc
import functools
import itertools
import types

import numpy as np

import pyxu.abc as pxa
import pyxu.info.deps as pxd
import pyxu.info.ptype as pxt
import pyxu.runtime as pxrt
import pyxu.util as pxu
from pyxu.info.plugin import _load_entry_points

__all__ = [
    "stack",
    "vstack",
    "hstack",
    "block_diag",
    "block",
    "coo_block",
]

__all__ = _load_entry_points(globals(), group="pyxu.opt.blocks", names=__all__)

# Arithmetic methods/attributes of [hstack, vstack, block_diag, coo_block]() are very similar. To avoid excessive
# copy/paste-ing the same implementation everywhere, these functions piggy-back on _COOBlock() for most computations.
# When a more efficient implementation is known however, it is overridden in [hstack, vstack, block_diag]() as needed.


def stack(
    ops: cabc.Sequence[pxt.OpT],
    axis: pxt.Integer,
    **kwargs,
) -> pxt.OpT:
    r"""
    Construct a stacked operator.

    A stacked-operator :math:`V: \mathbb{R}^{d} \to \mathbb{R}^{c}` is an operator containing (vertically or
    horizontally) blocks of smaller operators :math:`\{O_{1}, \ldots, O_{N}\}`.

    This is a convenience function around :py:func:`~pyxu.operator.hstack` and :py:func:`~pyxu.operator.vstack`.

    Parameters
    ----------
    ops: :py:class:`~collections.abc.Sequence` ( :py:attr:`~pyxu.info.ptype.OpT` )
        [op1(c1, d1), ..., opN(cN, dN)] operators to join.
    axis: Integer
        The axis along which operators will be joined, i.e.

        * 0: stack vertically (row-wise)
        * 1: stack horizontally (column-wise)
    kwargs
        Extra keyword-arguments forwarded to :py:class:`~pyxu.operator.blocks._COOBlock`.

    Returns
    -------
    op: OpT
        Stacked operator.

    See Also
    --------
    :py:func:`~pyxu.operator.hstack`,
    :py:func:`~pyxu.operator.vstack`.
    """
    axis = int(axis)
    assert axis in {0, 1}, f"axis: out-of-bounds axis '{axis}'."

    f = {0: vstack, 1: hstack}[axis]
    op = f(ops, **kwargs)
    return op


def vstack(
    ops: cabc.Sequence[pxt.OpT],
    **kwargs,
) -> pxt.OpT:
    r"""
    Construct a vertically-stacked operator.

    A vstacked-operator :math:`V: \mathbb{R}^{d} \to \mathbb{R}^{c_{1} + \cdots + c_{N}}` is an operator containing
    (vertically) blocks of smaller operators :math:`\{O_{1}: \mathbb{R}^{d} \to \mathbb{R}^{c_{1}}, \ldots, O_{N}:
    \mathbb{R}^{d} \to \mathbb{R}^{c_{N}}\}`, i.e.

    .. math::

       V
       =
       \left[
           \begin{array}{c}
               O_{1}  \\
               \vdots \\
               O_{N}  \\
           \end{array}
       \right]

    Parameters
    ----------
    ops: :py:class:`~collections.abc.Sequence` ( :py:attr:`~pyxu.info.ptype.OpT` )
        [op1(c1, d), ..., opN(cN, d)] operators to concatenate.
    kwargs
        Extra keyword-arguments forwarded to :py:class:`~pyxu.operator.blocks._COOBlock`.

    Returns
    -------
    op: OpT
        Vertically-stacked (c1+...+cN, d) operator.

    Notes
    -----
    * All sub-operator domains must have compatible shapes, i.e. all integer-valued ``dim`` s must be identical.

    See Also
    --------
    :py:func:`~pyxu.operator.hstack`,
    :py:func:`~pyxu.operator.stack`.
    """

    def op_expr(_) -> tuple:
        N_row = _._grid_shape[0]
        ops = [_._block[(r, 0)] for r in range(N_row)]
        return ("vstack", *ops)

    N_data = len(ops)
    op = _COOBlock(
        ops=(
            ops,  # data
            (
                tuple(range(N_data)),  # i
                [0] * N_data,  # j
            ),
        ),
        grid_shape=(N_data, 1),
        parallel=kwargs.get("parallel", False),
    ).op()
    op._expr = types.MethodType(op_expr, op)
    return op


def hstack(
    ops: cabc.Sequence[pxt.OpT],
    **kwargs,
) -> pxt.OpT:
    r"""
    Construct a horizontally-stacked operator.

    A hstacked-operator :math:`H: \mathbb{R}^{d_{1} + \cdots + d_{N}} \to \mathbb{R}^{c}` is an operator containing
    (horizontally) blocks of smaller operators :math:`\{O_{1}: \mathbb{R}^{d_{1}} \to \mathbb{R}^{c}, \ldots, O_{N}:
    \mathbb{R}^{d_{N}} \to \mathbb{R}^{c}\}`, i.e.

    .. math::

       H
       =
       \left[
           \begin{array}{ccc}
               O_{1} & \cdots & O_{N}
           \end{array}
       \right]

    Parameters
    ----------
    ops: :py:class:`~collections.abc.Sequence` ( :py:attr:`~pyxu.info.ptype.OpT` )
        [op1(c, d1), ..., opN(c, dN)] operators to concatenate.
    kwargs
        Extra keyword-arguments forwarded to :py:class:`~pyxu.operator.blocks._COOBlock`.

    Returns
    -------
    op: OpT
        Horizontally-stacked (c, d1+....+dN) operator.

    Notes
    -----
    * All sub-operator domains must have compatible shapes, i.e. all ``codim`` s must be identical.

    See Also
    --------
    :py:func:`~pyxu.operator.vstack`,
    :py:func:`~pyxu.operator.stack`.
    """

    def op_expr(_) -> tuple:
        N_col = _._grid_shape[1]
        ops = [_._block[(0, c)] for c in range(N_col)]
        return ("hstack", *ops)

    N_data = len(ops)
    op = _COOBlock(
        ops=(
            ops,  # data
            (
                [0] * N_data,  # i
                tuple(range(N_data)),  # j
            ),
        ),
        grid_shape=(1, N_data),
        parallel=kwargs.get("parallel", False),
    ).op()
    op._expr = types.MethodType(op_expr, op)
    return op


def block_diag(
    ops: cabc.Sequence[pxt.OpT],
    **kwargs,
) -> pxt.OpT:
    r"""
    Construct a block-diagonal operator.

    A block-diagonal operator :math:`D: \mathbb{R}^{d_{1} + \cdots + d_{N}} \to \mathbb{R}^{c_{1} + \cdots + c_{N}}` is
    an operator containing (diagonally) blocks of smaller operators :math:`\{O_{1}: \mathbb{R}^{d_{1}} \to
    \mathbb{R}^{c_{1}}, \ldots, O_{N}: \mathbb{R}^{d_{N}} \to \mathbb{R}^{c_{N}}\}`, i.e.

    .. math::

       D
       =
       \left[
           \begin{array}{ccc}
               O_{1} &        &       \\
                     & \ddots &       \\
                     &        & O_{N} \\
           \end{array}
       \right]

    Parameters
    ----------
    ops: :py:class:`~collections.abc.Sequence` ( :py:attr:`~pyxu.info.ptype.OpT` )
        [op1(c1, d1), ..., opN(cN, dN)] operators to concatenate.
    kwargs
        Extra keyword-arguments forwarded to :py:class:`~pyxu.operator.blocks._COOBlock`.

    Returns
    -------
    op: OpT
        Block-diagonal (c1+...+cN, d1+...+dN) operator.

    See Also
    --------
    :py:func:`~pyxu.operator.block`,
    :py:func:`~pyxu.operator.coo_block`.
    """

    def op_svdvals(_, **kwargs) -> pxt.NDArray:
        # op.svdvals(**kwargs) = [top|bottom-k]([op1.svdvals(**kwargs), ..., opN.svdvals(**kwargs)])
        if not _.has(pxa.Property.LINEAR):
            raise NotImplementedError

        k = kwargs.get("k", 1)
        which = kwargs.get("which", "LM")
        if which.upper() == "SM":
            D = _.__class__.svdvals(_, **kwargs)
        else:
            parts = [op.svdvals(**kwargs) for op in _._block.values()]
            xp = pxu.get_array_module(parts[0])
            D = xp.sort(xp.concatenate(parts, axis=0), axis=None)[-k:]
        return D

    @pxrt.enforce_precision(i=("arr", "damp"))
    def op_pinv(_, arr: pxt.NDArray, damp: pxt.Real, **kwargs) -> pxt.NDArray:
        # op.pinv(y, damp) = concatenate([op1.pinv(y1, damp), ..., opN.pinv(yN, damp)], axis=-1)
        if not _.has(pxa.Property.LINEAR):
            raise NotImplementedError

        parts = dict()
        for idx, op in _._block.items():
            offset = _._block_offset[idx][0]
            p = op.pinv(arr[..., offset : offset + op.codim], damp, **kwargs)
            parts[idx] = p

        xp = pxu.get_array_module(arr)
        parts = [parts[k] for k in sorted(parts.keys())]
        out = xp.concatenate(parts, axis=-1)
        return out

    @pxrt.enforce_precision()
    def op_trace(_, **kwargs) -> pxt.Real:
        # op.trace(**kwargs) = if [op1,...,opN] square
        #                          sum([op1.trace(**kwargs), ..., opN.trace(**kwargs)])
        #                      else
        #                          SquareOp.trace(**kwargs)
        if not _.has(pxa.Property.LINEAR_SQUARE):
            raise NotImplementedError

        if all(op.has(pxa.Property.LINEAR_SQUARE) for op in _._block.values()):
            parts = [op.trace(**kwargs) for op in _._block.values()]
            tr = sum(parts)
        else:  # default fallback
            tr = pxa.SquareOp.trace(_, **kwargs)
        return tr

    def op_expr(_) -> tuple:
        ops = [_._block[k] for k in sorted(_._block.keys())]
        return ("block_diag", *ops)

    N_data = len(ops)
    op = _COOBlock(
        ops=(
            ops,  # data
            (
                tuple(range(N_data)),  # i
                tuple(range(N_data)),  # j
            ),
        ),
        grid_shape=(N_data, N_data),
        parallel=kwargs.get("parallel", False),
    ).op()
    op.svdvals = types.MethodType(op_svdvals, op)
    op.pinv = types.MethodType(op_pinv, op)
    op.trace = types.MethodType(op_trace, op)
    op._expr = types.MethodType(op_expr, op)
    return op


def block(
    ops: cabc.Sequence[cabc.Sequence[pxt.OpT]],
    order: pxt.Integer,
    **kwargs,
) -> pxt.OpT:
    r"""
    Construct a (dense) block-defined operator.

    A block-defined operator is an operator containing blocks of smaller operators.  Blocks are stacked
    horizontally/vertically in a user-specified order to obtain the final shape.

    Parameters
    ----------
    ops: :py:class:`~collections.abc.Sequence` ( :py:class:`~collections.abc.Sequence` ( :py:attr:`~pyxu.info.ptype.OpT` ) )
        2D nested sequence of (ck, dk)-shaped operators.
    order: 0, 1
        Order in which the nested operators are specified/concatenated:

        * 0: concatenate inner-most blocks via :py:func:`~pyxu.operator.vstack`, then :py:func:`~pyxu.operator.hstack`.
        * 1: concatenate inner-most blocks via :py:func:`~pyxu.operator.hstack`, then :py:func:`~pyxu.operator.vstack`.
    kwargs
        Extra keyword-arguments forwarded to :py:class:`~pyxu.operator.blocks._COOBlock`.

    Returns
    -------
    op: OpT
        Block-defined operator. (See below for examples.)

    Notes
    -----
    * Each row/column may contain a different number of operators.

    Examples
    --------

    .. code::

       >>> block(
       ...    [
       ...     [A],        # ABEEGGG
       ...     [B, C, D],  # ACEEHHH
       ...     [E, F],     # ADFFHHH
       ...     [G, H],
       ...    ],
       ...    order=0,
       ... )

       >>> block(
       ...    [
       ...     [A, B, C, D],  # ABBCD
       ...     [E],           # EEEEE
       ...     [F, G],        # FFGGG
       ...    ],
       ...    order=1,
       ... )

    See Also
    --------
    :py:func:`~pyxu.operator.block_diag`,
    :py:func:`~pyxu.operator.coo_block`.
    """
    order = int(order)
    assert order in {0, 1}, f"order: out-of-bounds order '{order}'."

    inner = {0: vstack, 1: hstack}[order]
    outer = {0: hstack, 1: vstack}[order]

    op = outer([inner(row, **kwargs) for row in ops], **kwargs)
    return op


def coo_block(
    ops: tuple[
        cabc.Sequence[pxt.OpT],
        tuple[
            cabc.Sequence[pxt.Integer],
            cabc.Sequence[pxt.Integer],
        ],
    ],
    grid_shape: pxt.OpShape,
    *,
    parallel: bool = False,
) -> pxt.OpT:
    r"""
    Constuct a (possibly-sparse) block-defined operator in COOrdinate format.

    A block-defined operator is an operator containing blocks of smaller operators.  Blocks must align on a coarse grid,
    akin to the COO format used to define sparse arrays.

    Parameters
    ----------
    ops
        (data, (i, j)) sequences defining block placement, i.e.

        * `data[:]` are :py:attr:`~pyxu.info.ptype.OpT` instances, in any order.
        * `i[:]` are the row indices of the block entries on the coarse grid.
        * `j[:]` are the column indices of the block entries on the coarse grid.

    grid_shape: OpShape
        (M, N) shape of the coarse grid.

    parallel: bool
        If ``true`` , use Dask to evaluate the following methods:

        * ``.apply()``
        * ``.prox()``
        * ``.grad()``
        * ``.adjoint()``
        * ``.[co]gram().[apply,adjoint]()``

    Returns
    -------
    op: OpT
        Block-defined operator. (See below for examples.)

    Notes
    -----
    * Blocks on the same row/column must have the same ``codim`` / ``dim`` s.
    * Each row/column of the coarse grid **must** contain at least one entry.
    * ``parallel=True`` only parallelizes execution when inputs to listed methods are NUMPY arrays.

    .. Warning::

       When processing Dask inputs, or when ``parallel=True``, operators which are not thread-safe may produce incorrect
       results.  There is no easy way to ensure thread-safety at the level of :py:mod:`~pyxu.operator.blocks` without
       impacting performance of all operators involved.  Users are thus responsible for executing block-defined
       operators correctly, i.e., if thread-unsafe operators are involved, stick to NumPy/CuPy inputs.

    Examples
    --------

    .. code::

       >>> coo_block(
       ...     ([A(500,1000), B(1,1000), C(500,500), D(1,3)],  # data
       ...      [
       ...       [0, 1, 0, 2],  # i
       ...       [0, 0, 2, 1],  # j
       ...      ]),
       ...     grid_shape=(3, 3),
       ... )

       | coarse_idx |      0       |    1    |      2      |
       |------------|--------------|---------|-------------|
       |          0 | A(500, 1000) |         | C(500, 500) |
       |          1 | B(1, 1000)   |         |             |
       |          2 |              | D(1, 3) |             |
    """
    op = _COOBlock(
        ops=ops,
        grid_shape=grid_shape,
        parallel=parallel,
    ).op()
    return op


def _parallelize(func: cabc.Callable) -> cabc.Callable:
    # Parallelize execution of func() under conditions.
    #
    # * func() must be one of the arithmetic methods [apply,prox,grad,adjoint,pinv]()
    # * the `_parallel` attribute must be attached to the instance for it to parallelize execution
    #   over NUMPY inputs.

    @functools.wraps(func)
    def wrapper(*ARGS, **KWARGS):
        func_args = pxu.parse_params(func, *ARGS, **KWARGS)

        arr = func_args.get("arr", None)
        N = pxd.NDArrayInfo
        parallelize = ARGS[0]._parallel and (N.from_obj(arr) == N.NUMPY)

        # [2022.09.26] Sepand Kashani
        # Q: Why do we parallelize only for NUMPY inputs and not CUPY?
        # A: Given an arithmetic method `f`, there is no obligation (for DASK inputs) to satisfy the
        #    relation `f(arr).chunk_type == arr.chunk_type`.
        #    In particular the relationship does not hold for LinFunc.[grad, adjoint]() unless the
        #    user provides a special implementation for DASK inputs.
        #    Consequence: the sum() / xp.concatenate() instructions in _COOBlock() may:
        #    * fail due to array-type mismatch; or
        #    * induce unnecessary CPU<>GPU transfers.

        if parallelize:
            xp = N.DASK.module()
            func_args.update(arr=xp.array(arr, dtype=arr.dtype))
        else:
            pass

        out = func(**func_args)
        f = {True: pxu.compute, False: lambda _: _}[parallelize]
        return f(out)

    return wrapper


class _COOBlock:  # See coo_block() for a detailed description.
    def __init__(
        self,
        ops: tuple[
            cabc.Sequence[pxt.OpT],
            tuple[
                cabc.Sequence[pxt.Integer],
                cabc.Sequence[pxt.Integer],
            ],
        ],
        grid_shape: pxt.OpShape,
        parallel: bool,
    ):
        self._grid_shape = tuple(grid_shape)
        self._init_spec(ops)
        self._parallel = bool(parallel)

    def op(self) -> pxt.OpT:
        """
        Returns
        -------
        op: pxt.OpT
            Synthesized operator given inputs to :py:meth:`~pyxu.operator.blocks._COOBlock.__init__`.
        """
        blk = self._block
        if len(blk) == 1:
            _, op = blk.popitem()
        else:
            import pyxu.abc.arithmetic as arithmetic

            op = self._infer_op()
            op._block = self._block  # embed for introspection
            op._block_offset = self._block_offset  # embed for introspection
            op._grid_shape = self._grid_shape  # embed for introspection
            op._parallel = self._parallel  # embed for introspection
            for p in op.properties():
                for name in p.arithmetic_methods():
                    func = getattr(self.__class__, name)
                    setattr(op, name, types.MethodType(func, op))
            arithmetic.Rule._propagate_constants(op)
        return op

    def _init_spec(self, ops):
        """
        Transform input into standardized form.

        Returns
        -------
        block: (int, int) -> pxt.OpT
            coarse_grid index -> Operator at that location.
        block_offset: (int, int) -> (int, int)
            coarse_grid index -> (top-left) fine_grid index.
        """
        data, (i, j) = ops
        N_row, N_col, N_block = *self._grid_shape, len(data)
        msg = "Incorrect COO parametrization"
        assert N_block == len(i) == len(j), msg
        assert 0 < N_block <= N_row * N_col, msg
        assert 0 <= min(i) <= max(i) < N_row, msg
        assert 0 <= min(j) <= max(j) < N_col, msg

        # block dimensions are compatible.
        row = collections.defaultdict(list)  # row_idx -> [block]
        col = collections.defaultdict(list)  # col_idx -> [block]
        for d, _i, _j in zip(data, i, j):
            row[_i].append(d)
            col[_j].append(d)
        msg = lambda dim, idx, dom: f"All sub-operators on {dim} {idx} must have same {dom} size."
        for k, v in row.items():
            assert len({_.codim for _ in v}) == 1, msg("row", k, "codomain")
        for k, v in col.items():
            assert len({_.dim for _ in v}) == 1, msg("column", k, "domain")

        # no empty lines/columns in coarse grid.
        msg = lambda _: f"Coarse grid contains empty {_}: cannot infer fine-grid dimensions."
        assert len(row) == N_row, msg("rows")
        assert len(col) == N_col, msg("columns")

        # ---------------------------------------------------------------------
        # create block
        block = dict()  # coarse_grid(int, int) -> pxt.OpT
        for d, _i, _j in zip(data, i, j):
            block[(_i, _j)] = d.squeeze()
        self._block = block

        # create block_offset
        block_offset = dict()  # coarse_grid(int, int) -> fine_grid(int, int)
        row_offset = np.cumsum([row[k][0].codim for k in range(N_row)])
        col_offset = np.cumsum([col[k][0].dim for k in range(N_col)])
        for i in range(N_row):
            for j in range(N_col):
                _i = 0 if (i == 0) else row_offset[i - 1]
                _j = 0 if (j == 0) else col_offset[j - 1]
                block_offset[(i, j)] = _i, _j
        self._block_offset = block_offset

    def _infer_op(self) -> pxt.OpT:
        # Create the abstract COO-operator to which methods will be assigned.
        blk = self._block  # shorthand

        row = collections.defaultdict(list)
        col = collections.defaultdict(list)
        for (r, c), op in blk.items():
            row[r].append(op)
            col[c].append(op)
        N_row, N_col = len(row), len(col)
        op_codim = sum(ops[0].codim for ops in row.values())
        op_dim = sum(ops[0].dim for ops in col.values())

        P = pxa.Property
        properties = set.intersection(*[set(op.properties()) for op in blk.values()])
        if op_codim > 1:
            properties -= {
                P.FUNCTIONAL,
                P.PROXIMABLE,
                P.DIFFERENTIABLE_FUNCTION,
                P.QUADRATIC,
            }

        if all(  # block_diag case
            [
                N_row == N_col == len(blk),
                all((r, r) in blk for r in range(N_row)),
            ]
        ):
            pass
        elif op_codim == 1:  # hstack case: special treatment of quadratics
            if all(op.has(P.QUADRATIC) for op in blk.values()):
                properties.add(P.QUADRATIC)
            elif any(op.has(P.QUADRATIC) for op in blk.values()):
                # possible quadratic if all other terms are linear.
                non_quad = [op for op in blk.values() if not op.has(P.QUADRATIC)]
                if all(op.has(P.LINEAR) for op in non_quad):
                    properties.add(P.QUADRATIC)
        else:  # (1) vstack case or (2) arbitrary fill-in case => non-functionals
            properties &= {
                P.CAN_EVAL,
                P.DIFFERENTIABLE,
                P.LINEAR,
            }
            if (op_codim == op_dim) and (P.LINEAR in properties):
                properties.add(P.LINEAR_SQUARE)

        klass = pxa.Operator._infer_operator_type(properties)
        op = klass(shape=(op_codim, op_dim))
        return op

    @_parallelize
    @pxrt.enforce_precision(i="arr")
    def apply(self, arr: pxt.NDArray) -> pxt.NDArray:
        # compute blocks
        parts = dict()
        for idx, op in self._block.items():
            offset = self._block_offset[idx][1]
            p = op.apply(arr[..., offset : offset + op.dim])
            parts[idx] = p

        # row-oriented reduction (via +)
        rows = collections.defaultdict(list)
        for (r, c), p in parts.items():
            rows[r].append(p)
        cols = [sum(rows[r]) for r in range(len(rows))]

        # concatenate to super-column
        xp = pxu.get_array_module(arr)
        out = xp.concatenate(cols, axis=-1)
        return out

    def __call__(self, arr: pxt.NDArray) -> pxt.NDArray:
        return self.apply(arr)

    def estimate_lipschitz(self, **kwargs) -> pxt.Real:
        Ls_all = np.zeros(self._grid_shape)  # squared Lipschitz constant of each block.

        no_eval = "__rule" in kwargs
        if no_eval:
            for (r, c), op in self._block.items():
                Ls_all[r, c] = op.lipschitz**2
        elif self.has(pxa.Property.LINEAR):
            L = self.__class__.estimate_lipschitz(self, **kwargs)
            return L
        else:
            for (r, c), op in self._block.items():
                Ls_all[r, c] = op.estimate_lipschitz(**kwargs) ** 2

        # [non-linear case] Various upper bounds apply depending on how the blocks are organized:
        #   * vertical alignment: L**2 = sum(L_k**2)
        #   * horizontal alignment: L**2 = sum(L_k**2)
        #   * block-diagonal alignment: L**2 = max(L_k**2)
        #   * arbitrary 2D alignment: L**2 = sum(L_k**2)
        #     [obtained via vertical+horizontal composition (or vice-versa)]
        if np.allclose(Ls_all.sum(), Ls_all.diagonal().sum()):  # block-diag case
            L = np.sqrt(Ls_all.max())
        else:
            L = np.sqrt(Ls_all.sum())
        return L

    def _expr(self) -> tuple:
        class _Block(pxa.Operator):
            def __init__(self, idx: tuple[int, int], op: pxt.OpT):
                super().__init__(shape=op.shape)
                self._name = op._name
                self._idx = tuple(idx)
                self._op = op

            def _expr(self) -> tuple:
                head = "block[" + ", ".join(map(str, self._idx)) + "]"
                return (head, self._op)

        # head = coo_block[grid_shape]
        head = "coo_block[" + ", ".join(map(str, self._grid_shape)) + "]"

        # tail = sequence of sub-blocks, encapsulated in _Block to obtain hierarchical view.
        tail = [_Block(idx=k, op=op) for (k, op) in self._block.items()]
        return (head, *tail)

    @_parallelize
    @pxrt.enforce_precision(i=("arr", "tau"))
    def prox(self, arr: pxt.NDArray, tau: pxt.Real) -> pxt.NDArray:
        if not self.has(pxa.Property.PROXIMABLE):
            raise NotImplementedError

        parts = dict()
        for idx, op in self._block.items():
            offset = self._block_offset[idx][1]
            p = op.prox(arr=arr[..., offset : offset + op.dim], tau=tau)
            parts[idx] = p

        xp = pxu.get_array_module(arr)
        parts = [parts[(0, c)] for c in range(len(parts))]  # re-ordering
        out = xp.concatenate(parts, axis=-1)
        return out

    def _quad_spec(self):
        if not self.has(pxa.Property.QUADRATIC):
            raise NotImplementedError

        parts = dict()
        for idx, op in self._block.items():
            if op.has(pxa.Property.QUADRATIC):
                _Q, _c, _t = op._quad_spec()
            else:  # op is necessarily LINEAR
                from pyxu.operator import NullOp

                _Q = NullOp(shape=(op.dim, op.dim)).asop(pxa.PosDefOp)
                _c = op
                _t = 0
            parts[idx] = _Q, _c, _t

        parts = [parts[k] for k in sorted(parts.keys())]  # re-ordering
        Q, c, t = zip(*parts)
        return (block_diag(Q), hstack(c), sum(t))

    def jacobian(self, arr: pxt.NDArray) -> pxt.OpT:
        if not self.has(pxa.Property.DIFFERENTIABLE):
            raise NotImplementedError

        if self.has(pxa.Property.LINEAR):
            out = self
        else:
            data, i, j = [], [], []
            for (r, c), op in self._block.items():
                offset = self._block_offset[(r, c)][1]
                opJ = op.jacobian(arr[offset : offset + op.dim])

                i.append(r)
                j.append(c)
                data.append(opJ)

            out = _COOBlock(
                ops=(data, (i, j)),
                grid_shape=self._grid_shape,
                parallel=self._parallel,
            ).op()
        return out

    def estimate_diff_lipschitz(self, **kwargs) -> pxt.Real:
        if not self.has(pxa.Property.DIFFERENTIABLE):
            raise NotImplementedError

        dLs_all = np.zeros(self._grid_shape)  # squared diff-Lipschitz constant of each block.

        no_eval = "__rule" in kwargs
        if no_eval:
            for (r, c), op in self._block.items():
                dLs_all[r, c] = op.diff_lipschitz**2
        elif self.has(pxa.Property.QUADRATIC):
            dL = pxa.QuadraticFunc.estimate_diff_lipschitz(self, **kwargs)
            return dL
        elif self.has(pxa.Property.LINEAR):
            dL = 0
            return dL
        else:
            for (r, c), op in self._block.items():
                dLs_all[r, c] = op.estimate_diff_lipschitz(**kwargs) ** 2

        # Various upper bounds apply depending on how the blocks are organized:
        #   * vertical alignment: dL**2 = sum(dL_k**2)
        #   * horizontal alignment: dL**2 = sum(dL_k**2)
        #   * block-diagonal alignment: dL**2 = max(dL_k**2)
        #   * arbitrary 2D alignment: dL**2 = sum(dL_k**2)
        #     [obtained via vertical+horizontal composition (or vice-versa)]
        if np.allclose(dLs_all.sum(), dLs_all.diagonal().sum()):  # block-diag case
            dL = np.sqrt(dLs_all.max())
        else:
            dL = np.sqrt(dLs_all.sum())
        return dL

    @_parallelize
    @pxrt.enforce_precision(i="arr")
    def grad(self, arr: pxt.NDArray) -> pxt.NDArray:
        if not self.has(pxa.Property.DIFFERENTIABLE_FUNCTION):
            raise NotImplementedError

        parts = dict()
        for idx, op in self._block.items():
            offset = self._block_offset[idx][1]
            p = op.grad(arr[..., offset : offset + op.dim])
            parts[idx] = p

        xp = pxu.get_array_module(arr)
        parts = [parts[(0, c)] for c in range(len(parts))]  # re-ordering
        out = xp.concatenate(parts, axis=-1)
        return out

    @_parallelize
    @pxrt.enforce_precision(i="arr")
    def adjoint(self, arr: pxt.NDArray) -> pxt.NDArray:
        if not self.has(pxa.Property.LINEAR):
            raise NotImplementedError

        # compute blocks
        parts = dict()
        for idx, op in self._block.items():
            offset = self._block_offset[idx][0]
            p = op.adjoint(arr[..., offset : offset + op.codim])
            parts[idx[::-1]] = p

        # row-oriented reduction (via +)
        rows = collections.defaultdict(list)
        for (r, c), p in parts.items():
            rows[r].append(p)
        cols = [sum(rows[r]) for r in range(len(rows))]

        # concatenate to super-column
        xp = pxu.get_array_module(arr)
        out = xp.concatenate(cols, axis=-1)
        return out

    def asarray(self, **kwargs) -> pxt.NDArray:
        if not self.has(pxa.Property.LINEAR):
            raise NotImplementedError

        parts = dict()
        for idx, op in self._block.items():
            p = op.asarray(**kwargs)
            parts[idx] = p

        xp = kwargs.get("xp", pxd.NDArrayInfo.default().module())
        dtype = kwargs.get("dtype", pxrt.getPrecision().value)
        A = xp.zeros(self.shape, dtype=dtype)
        for idx, p in parts.items():
            r_o, c_o = self._block_offset[idx]  # offsets
            r_s, c_s = self._block[idx].shape  # spans
            A[r_o : r_o + r_s, c_o : c_o + c_s] = p
        return A

    def svdvals(self, **kwargs) -> pxt.NDArray:
        D = self.__class__.svdvals(self, **kwargs)
        return D

    @pxrt.enforce_precision(i=("arr", "damp"))
    def pinv(self, arr: pxt.NDArray, damp: pxt.Real, **kwargs) -> pxt.NDArray:
        out = self.__class__.pinv(self, arr=arr, damp=damp, **kwargs)
        return out

    def gram(self) -> pxt.OpT:
        if not self.has(pxa.Property.LINEAR):
            raise NotImplementedError

        blk = self._block  # shorthand
        N_row, N_col = self._grid_shape

        # Determine operator(s) which will occupy position (r,c) on coarse grid.
        ops = collections.defaultdict(list)  # coarse_grid(int, int) -> [pxt.OpT]
        for r, c in itertools.product(range(N_col), repeat=2):
            for k in range(N_row):
                if ((k, r) in blk) and ((k, c) in blk):
                    if r == c:
                        _op = blk[(k, r)].gram()
                    else:
                        _op = blk[(k, r)].T * blk[(k, c)]
                    ops[(r, c)].append(_op)

        # ops[(r,c)] should be reduced (via +) to form a single operator per (r,c)-entry.
        # It is inefficient however to chain so many operators together via AddRule().
        # G.[apply,adjoint]() are thus redefined to improve performance.
        data, i, j = [], [], []
        for (r, c), _ops in ops.items():
            klass = pxa.SelfAdjointOp if (r == c) else pxa.LinOp
            _op = klass(shape=_ops[0].shape)  # .[apply|adjoint]() overridden below.
            _op._ops = _ops  # embed for introspection

            @pxrt.enforce_precision(i="arr")
            def op_apply(_, arr: pxt.NDArray) -> pxt.NDArray:
                parts = [op.apply(arr) for op in _._ops]
                out = sum(parts)
                return out

            @pxrt.enforce_precision(i="arr")
            def op_adjoint(_, arr: pxt.NDArray) -> pxt.NDArray:
                parts = [op.adjoint(arr) for op in _._ops]
                out = sum(parts)
                return out

            _op.apply = types.MethodType(op_apply, _op)
            _op.adjoint = types.MethodType(op_adjoint, _op)

            data.append(_op)
            i.append(r)
            j.append(c)
        G = (
            _COOBlock(
                ops=(data, (i, j)),
                grid_shape=(N_col, N_col),
                parallel=self._parallel,
            )
            .op()
            .asop(pxa.SelfAdjointOp)
        )
        return G.squeeze()

    def cogram(self) -> pxt.OpT:
        if not self.has(pxa.Property.LINEAR):
            raise NotImplementedError

        blk = self._block  # shorthand
        N_row, N_col = self._grid_shape

        # Determine operator(s) which will occupy position (r,c) on coarse grid.
        ops = collections.defaultdict(list)  # coarse_grid(int, int) -> [pxt.OpT]
        for r, c in itertools.product(range(N_row), repeat=2):
            for k in range(N_col):
                if ((r, k) in blk) and ((c, k) in blk):
                    if r == c:
                        _op = blk[(r, k)].cogram()
                    else:
                        _op = blk[(r, k)] * blk[(c, k)].T
                    ops[(r, c)].append(_op)

        # ops[(r,c)] should be reduced (via +) to form a single operator per (r,c)-entry.
        # It is inefficient however to chain so many operators together via AddRule().
        # CG.[apply,adjoint]() are thus redefined to improve performance.
        data, i, j = [], [], []
        for (r, c), _ops in ops.items():
            klass = pxa.SelfAdjointOp if (r == c) else pxa.LinOp
            _op = klass(shape=_ops[0].shape)  # .[apply|adjoint]() overridden below.
            _op._ops = _ops  # embed for introspection

            @pxrt.enforce_precision(i="arr")
            def op_apply(_, arr: pxt.NDArray) -> pxt.NDArray:
                parts = [op.apply(arr) for op in _._ops]
                out = sum(parts)
                return out

            @pxrt.enforce_precision(i="arr")
            def op_adjoint(_, arr: pxt.NDArray) -> pxt.NDArray:
                parts = [op.adjoint(arr) for op in _._ops]
                out = sum(parts)
                return out

            _op.apply = types.MethodType(op_apply, _op)
            _op.adjoint = types.MethodType(op_adjoint, _op)

            data.append(_op)
            i.append(r)
            j.append(c)
        CG = (
            _COOBlock(
                ops=(data, (i, j)),
                grid_shape=(N_row, N_row),
                parallel=self._parallel,
            )
            .op()
            .asop(pxa.SelfAdjointOp)
        )
        return CG.squeeze()

    @pxrt.enforce_precision()
    def trace(self, **kwargs) -> pxt.Real:
        tr = self.__class__.trace(self, **kwargs)
        return tr

    def asloss(self, data: pxt.NDArray = None) -> pxt.OpT:
        msg = "asloss() is ambiguous for block-defined operators."
        raise NotImplementedError(msg)
